import base64

from assets.plugins.importers.plugin_file_importer import TEST_DOC, TestFileImporterPlugin

from steamship.data.file import File
from steamship.invocable import InvocableResponse
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.outputs.raw_data_plugin_output import RawDataPluginOutput

TEST_REQ = File()
TEST_PLUGIN_REQ = PluginRequest(data=TEST_REQ)
TEST_PLUGIN_REQ_DICT = TEST_PLUGIN_REQ.dict()


def _test_resp(res):
    assert isinstance(res, InvocableResponse)
    assert isinstance(res.data, RawDataPluginOutput)
    b64 = base64.b64encode(TEST_DOC.encode("utf-8")).decode("utf-8")
    assert res.data.data == b64


def test_importer():
    importer = TestFileImporterPlugin()
    res = importer.run(TEST_PLUGIN_REQ)
    _test_resp(res)

    # The endpoints take a kwargs block which is transformed into the appropriate JSON object
    res2 = importer.run_endpoint(**TEST_PLUGIN_REQ_DICT)
    _test_resp(res2)



from steamship_tests.utils.fixtures import get_steamship_client

from steamship import File
from steamship.base import TaskState


def test_file_upload_then_parse():
    steamship = get_steamship_client()

    a = File.create(
        steamship,
        content="This is a test.",
    )
    assert a.id is not None

    q1 = a.refresh()
    assert len(q1.blocks) == 0

    task = a.blockify(plugin_instance="markdown-blockifier-default-1.0")
    assert task is not None
    assert task.state == TaskState.waiting

    task.wait()
    assert task is not None
    assert task.state == TaskState.succeeded

    q1 = a.refresh()
    assert len(q1.blocks) == 1
    assert q1.blocks[0].text == "This is a test."

    b = File.create(
        steamship,
        content="""# Header

This is a test.""",
    )
    assert b.id is not None

    q1 = b.refresh()
    assert len(q1.blocks) == 0

    task = b.blockify(plugin_instance="markdown-blockifier-default-1.0")
    task.wait()

    q1 = b.refresh()
    assert len(q1.blocks) == 2
    assert q1.blocks[1].text == "This is a test."
    assert q1.blocks[0].text == "Header"

    a.delete()
    b.delete()



import pytest
from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import SteamshipError


def test_safe_loaded_package_with_bad_import():
    client = get_steamship_client()
    demo_package_path = PACKAGES_PATH / "safe_loaded_bad_import.pyignore"

    with pytest.raises(SteamshipError) as error:
        with deploy_package(client, demo_package_path) as (_, _, instance):
            pass  # Shouldn't get here!
    assert error is not None
    assert "No module named 'somethingthatclearlydoesnotexist'" in error.value.message



from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package
from steamship_tests.utils.fixtures import get_steamship_client


def test_returns_list():

    client = get_steamship_client()
    demo_package_path = PACKAGES_PATH / "returns_list.py"

    with deploy_package(client, demo_package_path) as (_, _, instance):
        response = instance.invoke("gimme_a_list")
        assert response is not None
        assert len(response) == 3
        assert response == ["here's", "a", "list"]



from enum import Enum

import pytest
from steamship_tests.utils.client import get_steamship_client
from steamship_tests.utils.random import random_name

from steamship import Steamship, Workspace
from steamship.utils.kv_store import KeyValueStore


class ExampleEnum(str, Enum):
    VALUE_1 = "value_1"
    VALUE_2 = "value_2"
    VALUE_3 = "value_3"


@pytest.mark.usefixtures("client")
def test_key_value_store(client: Steamship):
    """We can test the app like a regular python object!"""
    client = get_steamship_client()

    kv = KeyValueStore(client=client)
    kv.reset()

    # Getting empty key is None
    assert kv.get("FOO") is None

    key1 = "FOO"
    key2 = "BAR"

    # Set then Get
    value = {"a": 3}
    value_2 = {"hi": "there", "b": 5}
    value_3 = {"j": 9}

    # Set and get
    kv.set(key1, value)
    assert kv.get(key1) == value

    # Get again
    assert kv.get(key1) == value

    # Make a new one
    kv.set(key2, value_2)
    assert kv.get(key2) == value_2
    assert kv.get(key2) != value

    # List all the keys
    res = kv.items()
    assert len(res) == 2
    assert (key1, value) in res
    assert (key2, value_2) in res

    # List just the keys with val1
    res = kv.items(filter_keys=[key1])
    assert len(res) == 1
    assert (key1, value) in res
    assert (key2, value_2) not in res

    # Overwite the new one
    kv.set(key2, value_3)
    got = kv.get(key2)
    got_2 = kv.get(key2)
    got_3 = kv.get(key2)

    # List just the keys with val1
    res = kv.items(filter_keys=[key2])
    assert len(res) == 1
    assert (key1, value) not in res
    assert (key2, value_3) in res

    assert got == value_3
    assert got != value_2
    assert got != value
    assert got == got_2
    assert got == got_3

    # Delete a key
    kv.delete(key1)
    assert kv.get(key1) is None

    # But still others
    assert kv.get(key2) == value_3

    # Test enum
    kv.set(key1, {"k": ExampleEnum.VALUE_2.value})
    stat = kv.get(key1)["k"]
    assert isinstance(stat, str)
    stat_enum = ExampleEnum(stat)
    assert isinstance(stat_enum, ExampleEnum)
    assert stat_enum == ExampleEnum.VALUE_2

    # Delete all
    kv.reset()
    assert kv.get(key2) is None


@pytest.mark.usefixtures("client")
def test_kv_namespace_works(client: Steamship):
    client = get_steamship_client()

    kv1 = KeyValueStore(client=client, store_identifier="namespace1")
    kv2 = KeyValueStore(client=client, store_identifier="namespace2")

    kv1.reset()
    kv2.reset()

    # Getting empty key is None
    key = "key"
    assert kv1.get(key) is None
    assert kv2.get(key) is None

    val1 = {"a": 3}
    val2 = {"a": 4}

    # Set and get
    kv1.set(key, val1)
    assert kv1.get(key) == val1
    assert kv2.get(key) != val1
    assert kv2.get(key) is None

    kv2.set(key, val2)
    assert kv1.get(key) == val1
    assert kv2.get(key) == val2


def test_kv_multi_space_works():
    space_handle = random_name()
    client1 = get_steamship_client(workspace_handle=space_handle)

    space_handle2 = random_name()
    client2 = get_steamship_client(workspace_handle=space_handle2)

    kv1 = KeyValueStore(client=client1)
    kv2 = KeyValueStore(client=client2)

    kv1.reset()
    kv2.reset()

    # Getting empty key is None
    key = "key"
    assert kv1.get(key) is None
    assert kv2.get(key) is None

    val1 = {"a": 3}
    val2 = {"a": 4}

    # Set and get
    kv1.set(key, val1)
    assert kv1.get(key) == val1
    assert kv2.get(key) != val1
    assert kv2.get(key) is None

    kv2.set(key, val2)
    assert kv1.get(key) == val1
    assert kv2.get(key) == val2

    # Clean up
    Workspace(client=client1, id=client1.config.workspace_id).delete()
    Workspace(client=client2, id=client2.config.workspace_id).delete()



"""Collection of unit steamship_tests and integration steamship_tests for the steamship client."""
from pathlib import Path

ROOT_PATH = Path(__file__).parent.parent
VENV_PATH = ROOT_PATH / ".venv"
SRC_PATH = ROOT_PATH / "src"
TEST_PATH = Path(__file__).parent
TEST_ASSETS_PATH = TEST_PATH / "assets"
APPS_PATH = TEST_ASSETS_PATH / "packages"
PLUGINS_PATH = TEST_ASSETS_PATH / "plugins"



from typing import Any, Dict

from pydantic import BaseModel

from steamship.client.skills import Skill
from steamship.client.vendors import Vendor


class SkillVendorConfig(BaseModel):
    plugin_handle: str
    config: Dict[str, Any]


SKILL_TO_PROVIDER: Dict[Skill, Dict[Vendor, SkillVendorConfig]] = {
    Skill.ENTITIES: {
        Vendor.OneAI: SkillVendorConfig(
            plugin_handle="oneai-tagger",
            config={"skills": ["names", "numbers", "business-entities"]},
        )
    },
    Skill.SUMMARY: {
        Vendor.OneAI: SkillVendorConfig(
            plugin_handle="oneai-tagger", config={"skills": ["summarize"]}
        )
    },
    Skill.SENTIMENTS: {
        Vendor.OneAI: SkillVendorConfig(
            plugin_handle="oneai-tagger", config={"skills": ["sentiments"]}
        )
    },
    Skill.EMOTIONS: {
        Vendor.OneAI: SkillVendorConfig(
            plugin_handle="oneai-tagger", config={"skills": ["emotions"]}
        )
    },
    Skill.TOPICS: {
        Vendor.OneAI: SkillVendorConfig(
            plugin_handle="oneai-tagger", config={"skills": ["article-topics"]}
        ),
    },
    Skill.HIGHLIGHTS: {
        Vendor.OneAI: SkillVendorConfig(
            plugin_handle="oneai-tagger", config={"skills": ["highlights"]}
        )
    },
    Skill.KEYWORDS: {
        Vendor.OneAI: SkillVendorConfig(
            plugin_handle="oneai-tagger", config={"skills": ["keywords"]}
        )
    },
}



from steamship_tests import PLUGINS_PATH, TEST_ASSETS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import Block, File, MimeTypes, PluginInstance


def test_e2e_generator():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "generators" / "test_generator.py"

    with deploy_plugin(client, parser_path, "generator") as (
        plugin,
        version,
        instance,
    ):
        test_doc = "Yo! Banana boy!"
        res = instance.generate(text=test_doc, append_output_to_file=False)
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 1
        assert res.output.blocks[0].text == "!yob ananaB !oY"


def test_e2e_generator_with_existing_file():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "generators" / "test_generator.py"

    with deploy_plugin(client, parser_path, "generator") as (
        plugin,
        version,
        instance,
    ):

        # do an all-blocks test
        test_file = File.create(client, blocks=[Block(text="Yo! Banana boy!")])
        res = instance.generate(
            input_file_id=test_file.id, append_output_to_file=True, output_file_id=test_file.id
        )
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 1
        assert res.output.blocks[0].text == "!yob ananaB !oY"

        test_file.refresh()
        assert len(test_file.blocks) == 2
        assert test_file.blocks[1].text == "!yob ananaB !oY"

        # do a some-blocks test
        res = instance.generate(
            input_file_id=test_file.id,
            input_file_start_block_index=1,
            input_file_end_block_index=2,
            append_output_to_file=True,
            output_file_id=test_file.id,
        )
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 1
        assert res.output.blocks[0].text == "Yo! Banana boy!"

        test_file.refresh()
        assert len(test_file.blocks) == 3
        assert test_file.blocks[2].text == "Yo! Banana boy!"


def test_e2e_generator_runtime_options():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "generators" / "test_generator.py"

    with deploy_plugin(client, parser_path, "generator") as (
        plugin,
        version,
        instance,
    ):
        # do an all-blocks test
        test_file = File.create(client, blocks=[Block(text="Yo! Banana boy!")])
        res = instance.generate(
            input_file_id=test_file.id,
            append_output_to_file=True,
            output_file_id=test_file.id,
            options={"test": "yes"},
        )
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 2
        assert res.output.blocks[0].text == "!yob ananaB !oY"
        assert res.output.blocks[1].text == '{"test": "yes"}'


def test_generator_ephemeral_image_output():
    client = get_steamship_client()
    plugin_instance = PluginInstance.create(client, plugin_handle="test-image-generator")
    generate_task = plugin_instance.generate(text="This won't be used")

    generate_task.wait()
    assert generate_task.output is not None
    assert len(generate_task.output.blocks) == 1
    assert generate_task.output.blocks[0].content_url is not None
    data = generate_task.output.blocks[0].raw()
    assert data.decode("UTF-8") == "PRETEND THIS IS THE DATA OF AN IMAGE"


def test_e2e_generate_from_image():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "generators" / "test_image_to_text_generator.py"

    with deploy_plugin(client, parser_path, "generator") as (
        _,
        _,
        generator,
    ):
        test_file = File.create(client, content="")
        palm_tree_path = TEST_ASSETS_PATH / "palm_tree.png"

        with palm_tree_path.open("rb") as f:
            palm_bytes = f.read()
        block = Block.create(
            client, file_id=test_file.id, content=palm_bytes, mime_type=MimeTypes.PNG
        )

        fetched_bytes = block.raw()

        assert palm_bytes == fetched_bytes

        generate_task = generator.generate(input_file_id=test_file.id)
        generate_task.wait()
        result = generate_task.output.blocks[0].text
        assert result == "Found 1 image blocks and fetched data from 1"



from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package
from steamship_tests.utils.fixtures import get_steamship_client


def test_safe_loaded_hello_world():

    client = get_steamship_client()
    demo_package_path = PACKAGES_PATH / "safe_loaded_hello_world.py"

    with deploy_package(client, demo_package_path) as (_, _, instance):
        response = instance.invoke("greet")
        assert response is not None
        assert response == "Hello, Person"



import logging
import shutil
from pathlib import Path
from typing import Optional


def zip_folder(folder: Path, into_file: Optional[Path]) -> Path:
    """Zips a folder on disk to a co-located zip-file of the same name.

    The resulting zip file does not contain the enclosing folder name provided.
    It contains only the children of that folder as its root elements.
    """
    logging.info(f"Zipping: {folder}")
    shutil.make_archive(str(folder).rstrip("/"), "zip", folder)
    zip_path_str = str(folder).rstrip("/") + ".zip"
    logging.info(f"Zipped: {zip_path_str}")

    if into_file is None:
        return Path(zip_path_str)

    # Move the archive to the desired destination

    # Ensure the path to the desired extraction folder exists
    if not into_file.parent.exists():
        into_file.parent.mkdir(parents=True, exist_ok=True)

    shutil.move(zip_path_str, into_file)
    return into_file


def unzip_folder(zip_file: Path, into_folder: Optional[Path]) -> Path:
    """Unzips a folder on disk, returning the path to the new folder resulting."""
    logging.info(f"Unzipping: {zip_file}")
    if into_folder is None:
        into_folder = zip_file.with_suffix("")  # Strips the '.zip' suffix

    # Ensure the path to the desired extraction folder exists
    if not into_folder.parent.exists():
        into_folder.parent.mkdir(parents=True, exist_ok=True)

    shutil.unpack_archive(zip_file, into_folder, "zip")
    logging.info(f"Unzipped: {into_folder}")
    return Path(into_folder)



"""Objects for recording and reporting upon the introspected interface of a Steamship Package."""
import inspect
from enum import Enum
from typing import Dict, List, Optional, Union, get_args, get_origin

import steamship
from steamship import SteamshipError
from steamship.base.configuration import CamelModel
from steamship.utils.url import Verb


class ArgSpec(CamelModel):
    """An argument passed to a method."""

    # The name of the argument.
    name: str
    # The kind of the argument, reported by str(annotation) via the `inspect` library. E.g. <class 'int'>
    kind: str
    # Possible values, if the kind is an enum type
    values: Optional[List[str]]

    def __init__(self, name: str, parameter: inspect.Parameter):
        if name == "self":
            raise SteamshipError(
                message="Attempt to interpret the `self` object as a method parameter."
            )
        values = None
        if isinstance(parameter.annotation, type):
            if issubclass(parameter.annotation, Enum):
                values = [choice.value for choice in parameter.annotation]
        elif get_origin(parameter.annotation) is Union:
            args = get_args(parameter.annotation)
            # For now, only deal with the case where the Union is an Optional[Enum]
            if len(args) == 2 and type(None) in args:
                optional_arg = [t for t in args if t != type(None)][0]  # noqa: E721
                if issubclass(optional_arg, Enum):
                    values = [choice.value for choice in optional_arg]

        super().__init__(name=name, kind=str(parameter.annotation), values=values)

    def pprint(self, name_width: Optional[int] = None, prefix: str = "") -> str:
        """Returns a pretty printable representation of this argument."""
        width = name_width or len(self.name)
        ret = f"{prefix}{self.name.ljust(width)} - {self.kind}"
        return ret


class MethodSpec(CamelModel):
    """A method, callable remotely, on an object."""

    # The HTTP Path at which the method is callable.
    path: str

    # The HTTP Verb at which the method is callable. Defaults to POST
    verb: str

    # The return type. Reported by str(annotation) via the `inspect` library. E.g. <class 'int'>
    returns: str

    # The docstring of the method.
    doc: Optional[str] = None

    # The named arguments of the method. Positional arguments are not permitted.
    args: Optional[List[ArgSpec]] = None

    # Additional configuration around this endpoint.
    # Note: The actual type of this is Optional[Dict[str, Union[str, bool, int, float]]]
    #       But if Pydantic sees that, it attempts to force all values to be str, which is wrong.
    config: Optional[Dict] = None

    @staticmethod
    def clean_path(path: str = "") -> str:
        """Ensure that the path always starts with /, and at minimum must be at least /."""
        if not path:
            path = "/"
        elif path[0] != "/":
            path = f"/{path}"
        return path

    def __init__(
        self,
        cls: object,
        name: str,
        path: str = None,
        verb: Verb = Verb.POST,
        config: Dict[str, Union[str, bool, int, float]] = None,
    ):
        # Set the path
        if path is None and name is not None:
            path = f"/{name}"
        path = MethodSpec.clean_path(path)

        # Get the function on the class so that we can inspect it
        func = getattr(cls, name)
        sig = inspect.signature(func)

        # Set the return type
        returns = str(sig.return_annotation)

        # Set the docstring
        doc = func.__doc__

        # Set the arguments
        args = []
        for p in sig.parameters:
            if p == "self":
                continue
            args.append(ArgSpec(p, sig.parameters[p]))

        super().__init__(path=path, verb=verb, returns=returns, doc=doc, args=args, config=config)

    def pprint(self, name_width: Optional[int] = None, prefix: str = "  ") -> str:
        """Returns a pretty printable representation of this method."""

        width = name_width or len(self.path)
        ret = f"{self.verb.ljust(4)} {self.path.lstrip('/').ljust(width)} -> {self.returns}"
        if self.args:
            name_width = max([(len(arg.name) if arg.name else 0) for arg in self.args])
            for arg in self.args:
                arg_doc_string = arg.print(name_width, prefix)
                ret += f"\n{arg_doc_string}"
        return ret


class PackageSpec(CamelModel):
    """A package, representing a remotely instantiable service."""

    # The name of the package
    name: str

    # The docstring of the package
    doc: Optional[str] = None

    # The SDK version this package is deployed with
    sdk_version: str = steamship.__version__

    # The list of methods the package exposes remotely
    methods: Optional[List[MethodSpec]] = None

    def pprint(self, prefix: str = "  ") -> str:
        """Returns a pretty printable representation of this package."""
        underline = "=" * len(self.name)
        ret = f"{self.name}\n{underline}\n"
        if self.doc:
            ret += f"{self.doc}\n\n"
        else:
            ret += "\n"

        if self.methods:
            name_width = max([len(method.path) or 0 for method in self.methods])
            for method in self.methods:
                method_doc_string = method.pprint(name_width, prefix)
                ret += f"\n{method_doc_string}"
        return ret



from abc import ABC, abstractmethod

from steamship.invocable import InvocableResponse, post
from steamship.invocable.plugin_service import PluginRequest, PluginService
from steamship.plugin.inputs.file_import_plugin_input import FileImportPluginInput
from steamship.plugin.outputs.raw_data_plugin_output import RawDataPluginOutput


# Note!
# =====
#
# This is the PLUGIN IMPLEMENTOR's View of a File Importer.
#
# If you are using the Steamship Client, you probably want steamship.client.operations.file_importer instead
# of this file.
#
class FileImporter(PluginService[FileImportPluginInput, RawDataPluginOutput], ABC):
    @abstractmethod
    def run(
        self, request: PluginRequest[FileImportPluginInput]
    ) -> InvocableResponse[RawDataPluginOutput]:
        raise NotImplementedError()

    @post("import")
    def run_endpoint(self, **kwargs) -> InvocableResponse[RawDataPluginOutput]:
        """Exposes the File Importer's `run` operation to the Steamship Engine via the expected HTTP path POST /import"""
        return self.run(PluginRequest[FileImportPluginInput](**kwargs))



from __future__ import annotations

from typing import List

from steamship.base.model import CamelModel
from steamship.data.file import Block


class RawBlockAndTagPluginOutput(CamelModel):
    blocks: List[Block]



from __future__ import annotations

from typing import Any, Dict, Optional, Type

from pydantic import BaseModel, Field

from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import DeleteRequest, IdentifierRequest, Request
from steamship.data.workspace import Workspace
from steamship.utils.url import Verb


class CreatePackageInstanceRequest(Request):
    id: str = None
    package_id: str = None
    package_handle: str = None
    package_version_id: str = None
    package_version_handle: str = None
    handle: str = None
    fetch_if_exists: bool = None
    config: Dict[str, Any] = None
    workspace_id: str = None


class PackageInstance(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    handle: str = None
    package_id: str = None
    package_handle: Optional[str] = None
    user_handle: str = None
    package_version_id: str = None
    package_version_handle: Optional[str] = None
    user_id: str = None
    invocation_url: str = None
    config: Dict[str, Any] = None
    workspace_id: str = None
    workspace_handle: str = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["packageInstance"] if "packageInstance" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        package_id: str = None,
        package_handle: str = None,
        package_version_id: str = None,
        package_version_handle: str = None,
        handle: str = None,
        fetch_if_exists: bool = None,
        config: Dict[str, Any] = None,
    ) -> PackageInstance:
        req = CreatePackageInstanceRequest(
            handle=handle,
            package_id=package_id,
            package_handle=package_handle,
            package_version_id=package_version_id,
            package_version_handle=package_version_handle,
            fetch_if_exists=fetch_if_exists,
            config=config,
        )

        return client.post("package/instance/create", payload=req, expect=PackageInstance)

    def delete(self) -> PackageInstance:
        req = DeleteRequest(id=self.id)
        return self.client.post("package/instance/delete", payload=req, expect=PackageInstance)

    def load_missing_workspace_handle(self):
        if (
            self.client is not None
            and self.workspace_handle is None
            and self.workspace_id is not None
        ):
            # Get the workspaceHandle
            workspace = Workspace.get(self.client, id_=self.workspace_id)
            if workspace:
                self.workspace_handle = workspace.handle

    @staticmethod
    def get(client: Client, handle: str) -> PackageInstance:
        return client.post(
            "package/instance/get", IdentifierRequest(handle=handle), expect=PackageInstance
        )

    def invoke(
        self, path: str, verb: Verb = Verb.POST, timeout_s: Optional[float] = None, **kwargs
    ):
        self.load_missing_workspace_handle()
        if path[0] == "/":
            path = path[1:]

        return self.client.call(
            verb=verb,
            operation=f"/{self.workspace_handle or '_'}/{self.handle or '_'}/{path}",
            payload=kwargs,
            is_package_call=True,
            package_owner=self.user_handle,
            package_id=self.package_id,
            package_instance_id=self.id,
            as_background_task=False,
            timeout_s=timeout_s,
        )

    def full_url_for(self, path: str):
        return f"{self.invocation_url}{path}"



from __future__ import annotations

import json
from typing import Any, Dict, List, Optional, Type

from pydantic import BaseModel, Field

from steamship.base import Task
from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import Request
from steamship.base.response import Response
from steamship.data.plugin import HostingMemory, HostingTimeout


class CreatePluginVersionRequest(Request):
    plugin_id: str = None
    handle: str = None
    hosting_memory: Optional[HostingMemory] = None
    hosting_timeout: Optional[HostingTimeout] = None
    hosting_handler: str = None
    is_public: bool = None
    is_default: bool = None
    type: str = "file"
    # Note: this is a Dict[str, Any] but should be transmitted to the Engine as a JSON string
    config_template: str = None


class ListPluginVersionsRequest(Request):
    handle: str
    plugin_id: str


class ListPluginVersionsResponse(Response):
    plugins: List[PluginVersion]


class PluginVersion(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    plugin_id: str = None
    handle: str = None
    hosting_memory: Optional[HostingMemory] = None
    hosting_timeout: Optional[HostingTimeout] = None
    hosting_handler: str = None
    is_public: bool = None
    is_default: bool = None
    config_template: Dict[str, Any] = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["pluginVersion"] if "pluginVersion" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        handle: str,
        plugin_id: str = None,
        filename: str = None,
        filebytes: bytes = None,
        hosting_memory: Optional[HostingMemory] = None,
        hosting_timeout: Optional[HostingTimeout] = None,
        hosting_handler: str = None,
        is_public: bool = None,
        is_default: bool = None,
        config_template: Dict[str, Any] = None,
    ) -> Task[PluginVersion]:

        if filename is None and filebytes is None:
            raise Exception("Either filename or filebytes must be provided.")
        if filename is not None and filebytes is not None:
            raise Exception("Only either filename or filebytes should be provided.")

        if filename is not None:
            with open(filename, "rb") as f:
                filebytes = f.read()

        req = CreatePluginVersionRequest(
            handle=handle,
            plugin_id=plugin_id,
            hosting_memory=hosting_memory,
            hosting_timeout=hosting_timeout,
            hosting_handler=hosting_handler,
            is_public=is_public,
            is_default=is_default,
            config_template=json.dumps(config_template or {}),
        )

        task = client.post(
            "plugin/version/create",
            payload=req,
            file=("plugin.zip", filebytes, "multipart/form-data"),
            expect=PluginVersion,
        )

        task.wait()
        return task.output

    @staticmethod
    def list(
        client: Client, plugin_id: str = None, handle: str = None, public: bool = True
    ) -> ListPluginVersionsResponse:
        return client.post(
            f"plugin/version/{'public' if public else 'private'}",
            ListPluginVersionsRequest(handle=handle, plugin_id=plugin_id),
            expect=ListPluginVersionsResponse,
        )



from abc import ABC, abstractmethod

from steamship.invocable import InvocableResponse, post
from steamship.invocable.plugin_service import PluginRequest, PluginService
from steamship.plugin.inputs.raw_data_plugin_input import RawDataPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput

# Note!
# =====
#
# This is the PLUGIN IMPLEMENTOR's View of a Blockifier.
#
# If you are using the Steamship Client, you probably want steamship.client.operations.converter instead
# of this file.
#


class Blockifier(PluginService[RawDataPluginInput, BlockAndTagPluginOutput], ABC):
    @abstractmethod
    def run(
        self, request: PluginRequest[RawDataPluginInput]
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        raise NotImplementedError()

    @post("blockify")
    def run_endpoint(self, **kwargs) -> InvocableResponse[BlockAndTagPluginOutput]:
        """Exposes the Corpus Importer's `run` operation to the Steamship Engine via the expected HTTP path POST /import"""
        return self.run(PluginRequest[RawDataPluginInput].parse_obj(kwargs))



from __future__ import annotations

from typing import Any, Dict, Optional, Type

from pydantic import BaseModel

from steamship.base.model import CamelModel
from steamship.plugin.inputs.export_plugin_input import ExportPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput


class TrainingParameterPluginOutput(CamelModel):
    machine_type: Optional[str] = None
    training_epochs: int = None
    testing_holdout_percent: float = None
    test_split_seed: int = None
    training_params: Dict[str, Any] = None
    inference_params: Dict[str, Any] = None
    export_request: ExportPluginInput = None

    @staticmethod
    def from_input(input: TrainingParameterPluginInput) -> TrainingParameterPluginOutput:
        return TrainingParameterPluginOutput(
            export_request=input.export_plugin_input,
            training_epochs=input.training_epochs,
            testing_holdout_percent=input.testing_holdout_percent,
            test_split_seed=input.test_split_seed,
            training_params=input.training_params,
            inference_params=input.inference_params,
        )

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj["export_request"] = obj.get("exportPluginInput")
        return super().parse_obj(obj)



from typing import Any

from steamship.base.mime_types import MimeTypes
from steamship.data.file import File
from steamship.invocable import InvocableResponse
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput


def check_mime(d: dict, mime: str):
    assert d.get("http", {}).get("headers", {}).get("Content-Type") == mime


def check_type(d: dict, t: Any):
    data = d.get("data")
    assert isinstance(data, t)


def check_val(d: dict, val: Any):
    data = d.get("data")
    assert data == val


def test_text_response():
    r = InvocableResponse.from_obj("Hi there")
    d = r.dict()
    check_mime(d, MimeTypes.TXT)
    check_type(d, str)
    check_val(d, "Hi there")


def test_dict_response():
    r = InvocableResponse.from_obj({"a": 1})
    d = r.dict()
    check_mime(d, MimeTypes.JSON)
    check_type(d, dict)
    check_val(d, {"a": 1})


def test_resp_response():
    o = BlockAndTagPluginOutput(file=File(value="Foo", blocks=[], tags=[]))
    r = InvocableResponse(json=o)
    d = r.dict()
    check_mime(d, MimeTypes.JSON)
    check_type(d, dict)
    check_val(d, o.dict(by_alias=True))



from steamship.invocable import InvocableResponse, create_handler
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.file_importer import FileImporter
from steamship.plugin.inputs.file_import_plugin_input import FileImportPluginInput
from steamship.plugin.outputs.raw_data_plugin_output import RawDataPluginOutput


class TestFileImporterPythonErrorPlugin(FileImporter):
    def run(
        self, request: PluginRequest[FileImportPluginInput]
    ) -> InvocableResponse[RawDataPluginOutput]:
        raise Exception("TestFileImporterPythonErrorPlugin")


handler = create_handler(TestFileImporterPythonErrorPlugin)



from typing import Callable, Optional

import pytest
from assets.packages.configurable_hello_world import HelloWorld
from assets.packages.demo_package import TestPackage
from assets.packages.fancy_types import FancyTypes
from assets.packages.optional_params import OptionalParams


@pytest.mark.parametrize("invocable_handler", [TestPackage], indirect=True)
def test_package_spec(invocable_handler: Callable[[str, str, Optional[dict]], dict]):
    """Test that the handler returns the proper directory information"""
    rd = invocable_handler("GET", "/__dir__", {}).get("data")

    assert rd.get("doc") is None
    assert rd.get("methods") is not None
    assert len(rd.get("methods")) == 20

    saw_public = False

    for method in rd.get("methods"):
        if method.get("path") == "/greet" and method.get("verb") == "GET":
            assert method.get("config") is not None
            assert method.get("config").get("public") is True
            assert method.get("config").get("timeout") == 10
            assert method.get("config").get("identifier") == "foo"
            assert method.get("config").get("body") == 98.6
            assert method.get("config").get("not_there") is None

            saw_public = True
        elif "public" in method.get("path"):
            assert method.get("config") is not None
            assert method.get("config").get("public") is True
        else:
            assert method.get("config") == {}

    assert saw_public


@pytest.mark.parametrize("invocable_handler", [FancyTypes], indirect=True)
def test_package_spec_fancy_types(invocable_handler: Callable[[str, str, Optional[dict]], dict]):
    """Test that the handler returns the proper directory information"""
    rd = invocable_handler("GET", "/__dir__", {}).get("data")

    assert len(rd.get("methods")) == 2
    for method in rd.get("methods"):
        if method.get("path") == "/enum_route":
            values = method.get("args")[0].get("values")
            assert values is not None
            assert len(values) == 2
            assert values[0] == "value1"
            assert values[1] == "value2"
        if method.get("path") == "/long_string_route":
            assert method.get("args")[0].get("values") is None
            assert (
                method.get("args")[0].get("kind")
                == "<class 'steamship.invocable.paramater_types.longstr'>"
            )


@pytest.mark.parametrize("invocable_handler", [HelloWorld], indirect=True)
def test_package_spec_missing_configuration(
    invocable_handler: Callable[[str, str, Optional[dict]], dict]
):
    """Test that the handler returns the proper directory information, even for a configuration-required invocable."""
    rd = invocable_handler("GET", "/__dir__", {}).get("data")

    assert rd.get("doc") is None
    assert rd.get("methods") is not None
    assert len(rd.get("methods")) == 4


@pytest.mark.parametrize("invocable_handler", [OptionalParams], indirect=True)
def test_package_spec_optional_params(
    invocable_handler: Callable[[str, str, Optional[dict]], dict]
):
    """Test that the handler returns the proper directory information"""
    rd = invocable_handler("GET", "/__dir__", {}).get("data")

    assert len(rd.get("methods")) == 1
    method = rd.get("methods")[0]
    values = method.get("args")[0].get("values")
    assert values is not None
    assert len(values) == 2
    assert values[0] == "value1"
    assert values[1] == "value2"



from __future__ import annotations

from steamship.base.model import CamelModel


class FileImportPluginInput(CamelModel):
    value: str = None
    data: str = None
    url: str = None
    plugin_instance: str = None
    mime_type: str = None



import logging
from enum import Enum
from os import environ
from typing import Optional


class Verb(str, Enum):
    GET = "GET"
    POST = "POST"


def is_local(base: str) -> bool:
    """Check if we are running the client locally."""
    return any(
        local_base in base
        for local_base in ("localhost", "127.0.0.1", "0:0:0:0", "host.docker.internal", "/test:")
    )


def apply_localstack_url_fix(url: Optional[str]) -> Optional[str]:
    logging.debug(f"URL {url}")
    localstack_hostname = environ.get("LOCALSTACK_HOSTNAME")
    if url and localstack_hostname is not None and localstack_hostname != "localhost":
        for host in ["127.0.0.1", "host.docker.internal", "localstack"]:
            url = url.replace(host, localstack_hostname)
            logging.info(f"Replacing domain {host} in {url} with {localstack_hostname}")
    return url



import json
import logging
import sys
import time
from os import path
from typing import Optional

import click

import steamship
from steamship import Steamship, SteamshipError
from steamship.base.configuration import Configuration
from steamship.cli.deploy import (
    PackageDeployer,
    PluginDeployer,
    bundle_deployable,
    update_config_template,
)
from steamship.cli.manifest_init_wizard import manifest_init_wizard
from steamship.cli.requirements_init_wizard import requirements_init_wizard
from steamship.cli.ship_spinner import ship_spinner
from steamship.data.manifest import DeployableType, Manifest
from steamship.data.user import User


@click.group()
def cli():
    pass


def initialize(suppress_message: bool = False):
    logging.root.setLevel(logging.FATAL)
    if not suppress_message:
        click.echo(f"Steamship PYTHON cli version {steamship.__version__}")


@click.command()
def login():
    """Log in to Steamship, creating ~/.steamship.json"""
    initialize()
    click.echo("Logging into Steamship.")
    if sys.argv[1] == "login":
        if Configuration.default_config_file_has_api_key():
            overwrite = click.confirm(
                text="You already have an API key in your .steamship.json file.  Do you want to remove it and login?",
                default=False,
            )
            if not overwrite:
                sys.exit(0)
            Configuration.remove_api_key_from_default_config()

        # Carry on with login
        client = Steamship()
        user = User.current(client)
        click.secho(f"🚢🚢🚢 Hooray! You're logged in with user handle: {user.handle} 🚢🚢🚢", fg="green")


@click.command()
def ships():
    """Ship some ships"""
    initialize()
    click.secho("Here are some ships:", fg="cyan")
    with ship_spinner():
        time.sleep(5)
    click.secho()


@click.command()
def deploy():
    """Deploy the package or plugin in this directory"""
    initialize()
    client = None
    try:
        client = Steamship()
    except SteamshipError as e:
        click.secho(e.message, fg="red")
        click.get_current_context().abort()

    user = User.current(client)
    if path.exists("steamship.json"):
        manifest = Manifest.load_manifest()
    else:
        manifest = manifest_init_wizard(client)
        manifest.save()

    if not path.exists("requirements.txt"):
        requirements_init_wizard()

    deployable_type = manifest.type

    update_config_template(manifest)

    deployer = None
    if deployable_type == DeployableType.PACKAGE:
        deployer = PackageDeployer()
    elif deployable_type == DeployableType.PLUGIN:
        deployer = PluginDeployer()
    else:
        click.secho("Deployable must be of type package or plugin.", fg="red")
        click.get_current_context().abort()

    deployable = deployer.create_or_fetch_deployable(client, user, manifest)

    click.echo("Bundling content... ", nl=False)
    bundle_deployable(manifest)
    click.echo("Done. 📦")

    _ = deployer.create_version(client, manifest, deployable.id)

    thing_url = f"{client.config.web_base}{deployable_type}s/{manifest.handle}"
    click.echo(
        f"Deployment was successful. View and share your new {deployable_type} here:\n\n{thing_url}\n"
    )

    # Common error conditions:
    # - Package/plugin handle already taken. [handled; asks user for new]
    # - Version handle already deployed. [handled; asks user for new]
    # - Bad parameter configuration. [mitigated by deriving template from Config object]
    # - Package content fails health check (ex. bad import) [Error caught while checking config object]


@click.command()
@click.option(
    "--workspace",
    "-w",
    required=True,
    type=str,
    help="Workspace handle used for scoping logs request. All requests MUST be scoped by workspace.",
)
@click.option(
    "--offset",
    "-o",
    default=0,
    type=int,
    help="Starting index from sorted logs to return a chunk. Used for paging. Defaults to 0.",
)
@click.option(
    "--number",
    "-n",
    default=50,
    type=int,
    help="Number of logs to return in a single batch. Defaults to 50.",
)
@click.option(
    "--package",
    "-p",
    type=str,
    help="Package handle. Used to filter logs returend to a specific package (across all instances).",
)
@click.option(
    "--instance",
    "-i",
    type=str,
    help="Instance handle. Used to filter logs returned to a specific instance of a package.",
)
@click.option(
    "--version",
    "-v",
    type=str,
    help="Version handle. Used to filter logs returned to a specific version of a package.",
)
@click.option(
    "--path",
    "request_path",
    type=str,
    help="Path invoked by a client operation. Used to filter logs returned to a specific invocation path.",
)
def logs(
    workspace: str,
    offset: int,
    number: int,
    package: Optional[str] = None,
    instance: Optional[str] = None,
    version: Optional[str] = None,
    request_path: Optional[str] = None,
):
    initialize(suppress_message=True)
    client = None
    try:
        client = Steamship(workspace=workspace)
    except SteamshipError as e:
        raise click.UsageError(message=e.message)

    click.echo(json.dumps(client.logs(offset, number, package, instance, version, request_path)))


cli.add_command(login)
cli.add_command(deploy)
cli.add_command(deploy, name="it")
cli.add_command(ships)
cli.add_command(logs)


if __name__ == "__main__":
    deploy([])



from __future__ import annotations

from typing import Dict, List

from steamship.base.request import Request


class EmbedRequest(Request):
    docs: List[str]
    plugin_instance: str
    metadata: Dict = None



import hashlib
import json
from typing import Dict, List, Optional, Union

Metadata = Union[int, float, bool, str, List, Dict]


def str_to_metadata(s: str) -> Optional[Metadata]:
    if s is None:
        return None
    return json.loads(s)


def metadata_to_str(m: Metadata) -> Optional[str]:
    if m is None:
        return None
    return json.dumps(m)


def hash_dict(d: Dict) -> str:
    """Returns the MD5 hash of a dictionary."""
    dhash = hashlib.md5()  # noqa: S303
    # Sort arguments so so that the string representation is always the same.
    encoded = json.dumps(d, sort_keys=True).encode()
    dhash.update(encoded)
    return dhash.hexdigest()



from __future__ import annotations

import logging
import uuid
from contextlib import contextmanager
from typing import Any, Dict, Generator, List, Optional

from pydantic import BaseModel

from steamship.base.client import Client
from steamship.base.configuration import Configuration
from steamship.base.error import SteamshipError
from steamship.client.skill_to_provider import SKILL_TO_PROVIDER
from steamship.client.skills import Skill
from steamship.client.vendors import Vendor
from steamship.data.embeddings import EmbedAndSearchRequest, QueryResults
from steamship.data.package.package_instance import PackageInstance
from steamship.data.plugin.index_plugin_instance import EmbeddingIndexPluginInstance
from steamship.data.plugin.plugin_instance import PluginInstance
from steamship.data.plugin.prompt_generation_plugin_instance import PromptGenerationPluginInstance
from steamship.data.workspace import Workspace
from steamship.utils.metadata import hash_dict

_logger = logging.getLogger(__name__)


class Steamship(Client):
    """Steamship Python Client."""

    # Some plugin instances use special subclasses which provide helper methods and/or more complex
    # behavior than typical PluginInstance subclass permits. Examples are:
    #
    # - Embedding indices (which much coordinate both embedding taggers & vector indices)
    # - Prompt generators (which benefit from supporting, prompt-specific, methods)
    _PLUGIN_INSTANCE_SUBCLASS_OVERRIDES = {
        "prompt-generation-default": PromptGenerationPluginInstance,
        "prompt-generation-trainable-default": PromptGenerationPluginInstance,
        "gpt3": PromptGenerationPluginInstance,
        "gpt-3": PromptGenerationPluginInstance,
        "cerebrium": PromptGenerationPluginInstance,
        "embedding-index": EmbeddingIndexPluginInstance,
    }

    def __init__(
        self,
        api_key: str = None,
        api_base: str = None,
        app_base: str = None,
        web_base: str = None,
        workspace: str = None,
        fail_if_workspace_exists: bool = False,
        profile: str = None,
        config_file: str = None,
        config: Configuration = None,
        trust_workspace_config: bool = False,  # For use by lambda_handler; don't fetch the workspace
        **kwargs,
    ):
        super().__init__(
            api_key=api_key,
            api_base=api_base,
            app_base=app_base,
            web_base=web_base,
            workspace=workspace,
            fail_if_workspace_exists=fail_if_workspace_exists,
            profile=profile,
            config_file=config_file,
            config=config,
            trust_workspace_config=trust_workspace_config,
            **kwargs,
        )
        # We use object.__setattr__ here in order to bypass Pydantic's overloading of it (which would block this
        # set unless we were to add this as a field)
        object.__setattr__(self, "use", self._instance_use)
        object.__setattr__(self, "use_plugin", self._instance_use_plugin)

    def __repr_args__(self: BaseModel) -> Any:
        """Because of the trick we've done with `use` and `use_plugin`, we need to exclude these from __repr__
        otherwise we'll get an infinite recursion."""
        return [
            (key, value)
            for key, value in self.__dict__.items()
            if key != "use" and key != "use_plugin"
        ]

    def embed_and_search(
        self,
        query: str,
        docs: List[str],
        plugin_instance: str,
        k: int = 1,
    ) -> QueryResults:
        req = EmbedAndSearchRequest(query=query, docs=docs, plugin_instance=plugin_instance, k=k)
        return self.post(
            "plugin/instance/embeddingSearch",
            req,
            expect=QueryResults,
        )

    @staticmethod
    @contextmanager
    def temporary_workspace(**kwargs) -> Generator["Steamship", None, None]:
        """Create a client rooted in a temporary workspace that will be deleted after use."""
        # Create a new client and switch to a temporary workspace
        client = Steamship(**kwargs)
        temporary_handle = "temp-" + str(uuid.uuid4())
        client.switch_workspace(temporary_handle)

        # Safety check that we are now working form the new workspace.
        if client.config.workspace_handle != temporary_handle:
            raise SteamshipError(
                message=f"Attempted to switch to temporary workspace {temporary_handle} but the client claimed to be working from {client.config.workspace_handle}"
            )

        yield client

        # Safely delete the temporary workspace. Here we re-fetch the workspace using the temporary_handle
        # in case the user switched workspaces yet again upon the client.
        workspace = Workspace.get(client, handle=temporary_handle)
        if workspace.handle != temporary_handle:
            raise SteamshipError(
                message=f"Was about to delete temporary workspace {temporary_handle} but its handle is different: {workspace.handle}"
            )
        else:
            workspace.delete()

    @staticmethod
    def use(
        package_handle: str,
        instance_handle: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        version: Optional[str] = None,
        fetch_if_exists: bool = True,
        workspace_handle: Optional[str] = None,
        **kwargs,
    ) -> PackageInstance:
        """Creates/loads an instance of package `package_handle`.

        The instance is named `instance_handle` and located in the Workspace named `instance_handle`. If no
        `instance_handle` is provided, the default is `package_handle`.

        For example, one may write the following to always get back the same package instance, no matter how many
        times you run it, scoped into its own workspace:

        ```python
        instance = Steamship.use('package-handle', 'instance-handle')
        ```

        One may also write:

        ```python
        instance = Steamship.use('package-handle') # Instance will also be named `package-handle`
        ```

        If you wish to override the usage of a workspace named `instance_handle`, you can provide the `workspace_handle`
        parameter.
        """
        if instance_handle is None:
            instance_handle = package_handle
        kwargs["workspace"] = workspace_handle or instance_handle
        client = Steamship(**kwargs)
        return client._instance_use(
            package_handle=package_handle,
            instance_handle=instance_handle,
            config=config,
            version=version,
            fetch_if_exists=fetch_if_exists,
        )

    def _instance_use(
        self,
        package_handle: str,
        instance_handle: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        version: Optional[str] = None,
        fetch_if_exists: bool = True,
    ) -> PackageInstance:
        """Creates/loads an instance of package `package_handle`.

        The instance is named `instance_handle` and located in the workspace this client is anchored to.
        If no `instance_handle` is provided, the default is `package_handle`.
        """

        if instance_handle is None:
            if config is None:
                instance_handle = package_handle
            else:
                instance_handle = f"{package_handle}-{hash_dict(config)}"

        return PackageInstance.create(
            self,
            package_handle=package_handle,
            package_version_handle=version,
            handle=instance_handle,
            config=config,
            fetch_if_exists=fetch_if_exists,
        )

    @staticmethod
    def use_plugin(
        plugin_handle: str,
        instance_handle: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        version: Optional[str] = None,
        fetch_if_exists: bool = True,
        workspace_handle: Optional[str] = None,
        **kwargs,
    ) -> PluginInstance:
        """Creates/loads an instance of plugin `plugin_handle`.

        The instance is named `instance_handle` and located in the Workspace named `instance_handle`.
        If no `instance_handle` is provided, the default is `plugin_handle`.

        For example, one may write the following to always get back the same plugin instance, no matter how many
        times you run it, scoped into its own workspace:

        ```python
        instance = Steamship.use_plugin('plugin-handle', 'instance-handle')
        ```

        One may also write:

        ```python
        instance = Steamship.use('plugin-handle') # Instance will also be named `plugin-handle`
        ```
        """
        if instance_handle is None:
            instance_handle = plugin_handle
        kwargs["workspace"] = workspace_handle or instance_handle
        client = Steamship(**kwargs)
        return client._instance_use_plugin(
            plugin_handle=plugin_handle,
            instance_handle=instance_handle,
            config=config,
            version=version,
            fetch_if_exists=fetch_if_exists,
        )

    def use_skill(
        self,
        skill: Skill,
        provider: Optional[Vendor] = None,
        instance_handle: Optional[str] = None,
        fetch_if_exists: Optional[bool] = True,
    ) -> PluginInstance:

        if skill not in SKILL_TO_PROVIDER:
            raise SteamshipError(
                f"Unsupported skill provided. "
                f"Use one of our supported skills: {','.join(SKILL_TO_PROVIDER)}"
            )

        if provider and provider not in SKILL_TO_PROVIDER[skill]:
            raise SteamshipError(
                f"The provider {provider} has no support for the skill {skill}."
                f"Use one of the providers that support your skill: "
                f"{','.join(SKILL_TO_PROVIDER[skill])}"
            )

        plugin_setup = (
            SKILL_TO_PROVIDER[skill][provider]
            if provider
            else list(SKILL_TO_PROVIDER[skill].values())[0]
        )
        return self._instance_use_plugin(
            plugin_handle=plugin_setup["plugin_handle"],
            instance_handle=instance_handle,
            config=plugin_setup["config"],
            fetch_if_exists=fetch_if_exists,
        )

    def _instance_use_plugin(
        self,
        plugin_handle: str,
        instance_handle: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        version: Optional[str] = None,
        fetch_if_exists: Optional[bool] = True,
    ) -> PluginInstance:
        """Creates/loads an instance of plugin `plugin_handle`.

        The instance is named `instance_handle` and located in the workspace this client is anchored to.
        If no `instance_handle` is provided, the default is `plugin_handle`.
        """

        if instance_handle is None:
            if config is None:
                instance_handle = plugin_handle
            else:
                instance_handle = f"{plugin_handle}-{hash_dict(config)}"

        if plugin_handle in Steamship._PLUGIN_INSTANCE_SUBCLASS_OVERRIDES:
            return Steamship._PLUGIN_INSTANCE_SUBCLASS_OVERRIDES[plugin_handle].create(
                self,
                plugin_handle=plugin_handle,
                plugin_version_handle=version,
                handle=instance_handle,
                config=config,
                fetch_if_exists=fetch_if_exists,
            )

        return PluginInstance.create(
            self,
            plugin_handle=plugin_handle,
            plugin_version_handle=version,
            handle=instance_handle,
            config=config,
            fetch_if_exists=fetch_if_exists,
        )

    def get_workspace(self) -> Workspace:
        # We should probably add a hard-coded way to get this. The client in a Steamship Plugin/App comes
        # pre-configured with an API key and the Workspace in which this client should be operating.
        # This is a way to load the model object for that workspace.
        logging.info(
            f"get_workspace() called on client with config workspace {self.config.workspace_handle}/{self.config.workspace_id}"
        )
        workspace = Workspace.get(
            self, id_=self.config.workspace_id, handle=self.config.workspace_handle
        )
        if not workspace:
            logging.error("Unable to get workspace.")
            raise SteamshipError(
                message="Error while retrieving the Workspace associated with this client config.",
                internal_message=f"workspace_id={self.config.workspace_id} workspace_handle={self.config.workspace_handle}",
            )
        logging.info(f"Got workspace: {workspace.handle}/{workspace.id}")
        return workspace



from steamship import Block, File, Tag
from steamship.data import GenerationTag, TagKind, TagValueKey
from steamship.invocable import InvocableResponse, create_handler
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.tagger import Tagger


class TestPromptGeneratorPlugin(Tagger):
    def run(
        self, request: PluginRequest[BlockAndTagPluginInput]
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        """Merely returns the prompt."""
        request_file = request.data.file
        output = BlockAndTagPluginOutput(file=File(id=request_file.id), tags=[])
        for block in request.data.file.blocks:
            text = block.text
            tags = [
                Tag(
                    kind=TagKind.GENERATION,
                    name=GenerationTag.PROMPT_COMPLETION,
                    value={TagValueKey.STRING_VALUE: text},
                )
            ]
            output_block = Block(id=block.id, tags=tags)
            output.file.blocks.append(output_block)

        return InvocableResponse(data=output)


handler = create_handler(TestPromptGeneratorPlugin)



import json
from typing import List

from steamship import Block
from steamship.invocable import InvocableResponse
from steamship.plugin.generator import Generator
from steamship.plugin.inputs.raw_block_and_tag_plugin_input import RawBlockAndTagPluginInput
from steamship.plugin.outputs.raw_block_and_tag_plugin_output import RawBlockAndTagPluginOutput
from steamship.plugin.request import PluginRequest


class TestGenerator(Generator):
    def run(
        self, request: PluginRequest[RawBlockAndTagPluginInput]
    ) -> InvocableResponse[RawBlockAndTagPluginOutput]:
        result: List[Block] = []

        for block in request.data.blocks:
            result.append(Block(text=block.text[::-1]))

        if request.data.options is not None:
            result.append(Block(text=json.dumps(request.data.options)))

        return InvocableResponse(data=RawBlockAndTagPluginOutput(blocks=result))



import re

import click
from click import BadParameter

from steamship import Steamship
from steamship.data.manifest import Manifest, PluginConfig, SteamshipRegistry
from steamship.data.user import User


def validate_handle(handle: str) -> str:
    if re.fullmatch(r"[a-z\-]+", handle) is not None:
        return handle
    else:
        raise BadParameter("Handle must only include lowercase letters and -")


def validate_version_handle(handle: str) -> str:
    if re.fullmatch(r"[a-z0-9\-.]+", handle) is not None:
        return handle
    else:
        raise BadParameter("Handle must only include lowercase letters, numbers, . and -")


def manifest_init_wizard(client: Steamship):
    click.secho(
        "It looks like you don't yet have a steamship.json to deploy. Let's create one.",
        fg="cyan",
    )

    deployable_type = click.prompt(
        "Is this a package or a plugin?",
        default="package",
        type=click.Choice(["package", "plugin"]),
        show_choices=False,
    )

    handle = click.prompt(
        f"What handle would you like to use for your {deployable_type}? Valid characters are a-z and -",
        value_proc=validate_handle,
    )

    # TODO: claim the handle right here!

    version_handle = "0.0.1"

    plugin_detail = None
    if deployable_type == "plugin":
        plugin_type = click.prompt(
            "What type of plugin is this?",
            default="tagger",
            type=click.Choice(
                ["tagger", "blockifier", "exporter", "fileImporter", "corpusImporter", "generator"]
            ),
            show_choices=True,
        )
        if plugin_type == "tagger":
            trainable = click.confirm("Is the plugin trainable?", default=False)
        else:
            trainable = False
        plugin_detail = PluginConfig(isTrainable=trainable, type=plugin_type)

    public = click.confirm(f"Do you want this {deployable_type} to be public?", default=True)

    user = User.current(client)

    author = click.prompt("How should we list your author name?", default=user.handle)

    tagline = None
    author_github = None
    if public:
        tagline = click.prompt(f"Want to give the {deployable_type} a tagline?", default="")
        author_github = click.prompt(
            "If you'd like this associated with your github account, please your github username",
            default="",
        )

    tag_string = click.prompt(
        f"Want to give the {deployable_type} some tags? (comma separated)", default="Prompt API"
    )
    tags = [tag.strip() for tag in tag_string.split(",")]

    return Manifest(
        type=deployable_type,
        handle=handle,
        version=version_handle,
        description="",
        author=author,
        public=public,
        plugin=plugin_detail,
        build_config={"ignore": ["tests", "examples"]},
        configTemplate={},
        steamshipRegistry=SteamshipRegistry(
            tagline=tagline, authorGithub=author_github, authorName=author, tags=tags
        ),
    )



import pytest
from steamship_tests import PLUGINS_PATH
from steamship_tests.client.operations.test_tag_file import tag_file
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import SteamshipError


def test_e2e_tagger():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "taggers" / "plugin_parser.py"
    # TODO (enias): Use Enum for plugin type
    with deploy_plugin(client, parser_path, "tagger") as (
        plugin,
        version,
        instance,
    ):
        test_doc = "Hi there"
        res = instance.tag(doc=test_doc)
        res.wait()
        assert res.output is not None
        assert len(res.output.file.blocks) == 1
        assert res.output.file.blocks[0].text == test_doc

        # Let's try it on a file. This is the same test we run on the Swift test parser.
        # Since the python test parser is implemented to behave the same, we can reuse it!
        tag_file(client, instance.handle)


def test_e2e_tagger_bad_import():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "taggers" / "plugin_tagger_bad_import.pyignore"
    # TODO (enias): Use Enum for plugin type

    with pytest.raises(SteamshipError) as error:
        with deploy_plugin(client, parser_path, "tagger") as (
            plugin,
            version,
            instance,
        ):
            pass  # Shouldn't get here!

    assert error is not None
    assert "No module named 'somethingthatclearlydoesnotexist'" in error.value.message



from typing import Optional, Type, Union

from steamship import Tag
from steamship.data import TagValueKey
from steamship.invocable import Config, InvocableResponse, create_handler
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.tagger import Tagger


class TestParserPlugin(Tagger):
    # For testing; mirrors TestConfigurableTagger in Swift

    class TestParserConfig(Config):
        tag_kind: str
        tag_name: str
        number_value: Optional[Union[int, float]] = None
        # TODO: Check to see if python and/or swift removes False from obj.
        # If this is non-optional, the typecheck fails despite the fact that the test passes
        # in a value of false....
        boolean_value: Optional[bool] = False
        string_value: Optional[str] = ""

    config: TestParserConfig

    @classmethod
    def config_cls(cls) -> Type[Config]:
        return cls.TestParserConfig

    def run(
        self, request: PluginRequest[BlockAndTagPluginInput]
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        tag_kind = self.config.tag_kind
        tag_name = self.config.tag_name
        tag_value = {
            TagValueKey.NUMBER_VALUE: self.config.number_value,
            TagValueKey.BOOL_VALUE: self.config.boolean_value,
            TagValueKey.STRING_VALUE: self.config.string_value,
        }

        if request.data is not None:
            file = request.data.file
            tag = Tag(kind=tag_kind, name=tag_name, value=tag_value)
            if file.tags:
                file.tags.append(tag)
            else:
                file.tags = [tag]
            return InvocableResponse(data=BlockAndTagPluginOutput(file=file))


handler = create_handler(TestParserPlugin)



import logging

from steamship.data.workspace import SignedUrl
from steamship.invocable import InvocableResponse, PackageService, post
from steamship.utils.signed_urls import upload_to_signed_url, url_to_bytes

TEST_STRING = "A happy little test string"


class SignedURLPackage(PackageService):
    @post("writeReadSignedURL")
    def write_read_signed_url(self) -> InvocableResponse:
        logging.info("Starting invocation of signed url test package")
        workspace = self.client.get_workspace()

        filepath = "testPath"
        write_signed_url = workspace.create_signed_url(
            SignedUrl.Request(
                bucket=SignedUrl.Bucket.PLUGIN_DATA,
                filepath=filepath,
                operation=SignedUrl.Operation.WRITE,
            )
        ).signed_url

        logging.info(f"Got signed url for writing: {write_signed_url}")

        upload_to_signed_url(write_signed_url, bytes(TEST_STRING, "utf-8"))
        logging.info("Uploaded to signed url successfully.")

        read_signed_url = workspace.create_signed_url(
            SignedUrl.Request(
                bucket=SignedUrl.Bucket.PLUGIN_DATA,
                filepath=filepath,
                operation=SignedUrl.Operation.READ,
            )
        ).signed_url

        logging.info(f"Got signed url for reading: {read_signed_url}")

        result_bytes = url_to_bytes(read_signed_url)
        result_text = result_bytes.decode("utf-8")

        logging.info(f"Read from signed url: {result_text}")
        return InvocableResponse(string=result_text)

    @post("readSignedURL")
    def read_signed_url(self, filepath: str) -> InvocableResponse:
        logging.info("Starting invocation of signed url test package; read only")
        workspace = self.client.get_workspace()

        read_signed_url = workspace.create_signed_url(
            SignedUrl.Request(
                bucket=SignedUrl.Bucket.PLUGIN_DATA,
                filepath=filepath,
                operation=SignedUrl.Operation.READ,
            )
        ).signed_url

        logging.info(f"Got signed url for reading: {read_signed_url}")

        result_bytes = url_to_bytes(read_signed_url)
        result_text = result_bytes.decode("utf-8")

        logging.info(f"Read from signed url: {result_text}")
        return InvocableResponse(string=result_text)



import logging

from steamship.invocable import Invocable, InvocableResponse, get, post


class HelloWorld(Invocable):
    @post("greet")
    def greet(self, name: str = "Person") -> InvocableResponse:
        logging.info("Invoked greet; internal to helloworld")
        return InvocableResponse(string=f"Hello, {name}")

    @get("workspace")
    def workspace(self) -> InvocableResponse:
        return InvocableResponse(string=self.client.config.workspace_id)



from steamship import Block, Tag
from steamship.invocable import InvocableResponse


def test_camel_all_levels():
    tag = Tag(kind="kind", name="name", start_idx=0, end_idx=10)
    tagd = tag.dict(by_alias=True)

    assert tagd.get("startIdx") == 0
    assert tagd.get("endIdx") == 10

    block = Block(tags=[tag])
    blockd = block.dict(by_alias=True)

    tags = blockd.get("tags")
    assert tags is not None
    assert len(tags) == 1
    tagd2 = tags[0]

    assert tagd2.get("startIdx") == 0
    assert tagd2.get("endIdx") == 10

    ir = InvocableResponse.from_obj(block)

    tags_ir = ir.data.get("tags")
    assert tags_ir is not None
    assert len(tags_ir) == 1
    tagd3 = tags_ir[0]

    assert tagd3.get("startIdx") == 0
    assert tagd3.get("endIdx") == 10



import importlib
import inspect
import json
import logging
import sys
import traceback
import uuid
from http import HTTPStatus
from os import environ
from typing import Callable, Dict, Type

from fluent import asynchandler as fluenthandler
from fluent.handler import FluentRecordFormatter

from steamship import Configuration
from steamship.base import SteamshipError
from steamship.client import Steamship
from steamship.data.workspace import SignedUrl
from steamship.invocable import Invocable, InvocableRequest, InvocableResponse, InvocationContext
from steamship.utils.signed_urls import upload_to_signed_url


def encode_exception(obj):
    """When logging an exception ex: logging.exception(some_error), the exception must be turned into a string
    so that it is accepted by elasticsearch"""
    if isinstance(obj, SteamshipError):
        return json.dumps(obj.to_dict())
    if isinstance(obj, Exception):
        return f"exception_class: {type(obj).__name__}, args: {obj.args}"
    return obj


def internal_handler(  # noqa: C901
    invocable_cls_func: Callable[[], Type[Invocable]],
    event: Dict,
    client: Steamship,
    invocation_context: InvocationContext,
) -> InvocableResponse:

    try:
        request = InvocableRequest.parse_obj(event)
    except SteamshipError as se:
        logging.exception(se)
        return InvocableResponse.from_obj(se)
    except Exception as ex:
        logging.exception(ex)
        return InvocableResponse.error(
            code=HTTPStatus.INTERNAL_SERVER_ERROR,
            message="Plugin/App handler was unable to parse inbound request.",
            exception=ex,
        )

    if request and request.invocation:
        error_prefix = (
            f"[ERROR - {request.invocation.http_verb} {request.invocation.invocation_path}] "
        )
    else:
        error_prefix = "[ERROR - ?VERB ?PATH] "

    if request.invocation.invocation_path == "/__dir__":
        # Return the DIR result without (1) Constructing invocable_cls or (2) Parsing its config (in the constructor)
        try:
            cls = invocable_cls_func()
            return InvocableResponse(json=cls.__steamship_dir__(cls))
        except SteamshipError as se:
            logging.exception(se)
            return InvocableResponse.from_obj(se)
        except Exception as ex:
            logging.exception(ex)
            return InvocableResponse.error(
                code=HTTPStatus.INTERNAL_SERVER_ERROR,
                prefix=error_prefix,
                message="Unable to initialize package/plugin.",
                exception=ex,
            )

    try:
        invocable = invocable_cls_func()(
            client=client, config=request.invocation.config, context=invocation_context
        )
    except SteamshipError as se:
        logging.exception(se)
        return InvocableResponse.from_obj(se)
    except Exception as ex:
        logging.exception(ex)
        return InvocableResponse.error(
            code=HTTPStatus.INTERNAL_SERVER_ERROR,
            prefix=error_prefix,
            message="Unable to initialize package/plugin.",
            exception=ex,
        )

    if not invocable:
        return InvocableResponse.error(
            code=HTTPStatus.INTERNAL_SERVER_ERROR,
            prefix=error_prefix,
            message="Unable to construct package/plugin for invocation.",
        )

    try:
        response = invocable(request)
        return InvocableResponse.from_obj(response)
    except SteamshipError as se:
        logging.exception(se)
        se.message = f"{error_prefix}{se.message}"
        return InvocableResponse.from_obj(se)
    except Exception as ex:
        logging.exception(ex)
        return InvocableResponse.error(
            code=HTTPStatus.INTERNAL_SERVER_ERROR,
            prefix=error_prefix,
            exception=ex,
        )


def handler(internal_handler, event: Dict, _: Dict = None) -> dict:  # noqa: C901
    logging_config = event.get("loggingConfig")

    if logging_config is None:
        return InvocableResponse.error(
            code=HTTPStatus.INTERNAL_SERVER_ERROR,
            message="Plugin/App handler did not receive a remote logging config.",
        ).dict(by_alias=True)

    logging_host = logging_config.get("loggingHost")
    logging_port = logging_config.get("loggingPort")

    logging.basicConfig(level=logging.INFO)
    logging_handler = None

    invocation_context_dict = event.get("invocationContext")
    if invocation_context_dict is None:
        return InvocableResponse.error(
            code=HTTPStatus.INTERNAL_SERVER_ERROR,
            message="Plugin/App handler did not receive an invocation context.",
        ).dict(by_alias=True)

    invocation_context = InvocationContext.parse_obj(invocation_context_dict)
    # These log statements intentionally go to the logging handler pre-remote attachment, to debug logging configuration issues
    logging.info(f"Logging host: {logging_host} Logging port: {logging_port}")
    logging.info(f"Invocation context: {invocation_context}")

    if (
        logging_host != "none"
    ):  # Key off the string none, not 'is None', to avoid config errors where remote host isn't passed
        # Configure remote logging
        if logging_host is None:
            return InvocableResponse.error(
                code=HTTPStatus.INTERNAL_SERVER_ERROR,
                message="Plugin/App handler did receive a remote logging config, but it did not include a loggingHost.",
            ).dict(by_alias=True)

        if logging_port is None:
            return InvocableResponse.error(
                code=HTTPStatus.INTERNAL_SERVER_ERROR,
                message="Plugin/App handler did receive a remote logging config, but it did not include a loggingPort.",
            ).dict(by_alias=True)

        custom_format = {
            "level": "%(levelname)s",
            "host": "%(hostname)s",
            "where": "%(module)s.%(filename)s.%(funcName)s:%(lineno)s",
            "type": "%(levelname)s",
            "stack_trace": "%(exc_text)s",
            "component": "package-plugin-lambda",
            "userId": invocation_context.user_id,
            "workspaceId": invocation_context.workspace_id,
            "tenantId": invocation_context.tenant_id,
            "invocableHandle": invocation_context.invocable_handle,
            "invocableVersionHandle": invocation_context.invocable_version_handle,
            "invocableInstanceHandle": invocation_context.invocable_instance_handle,
            "invocableType": invocation_context.invocable_type,
            "invocableOwnerId": invocation_context.invocable_owner_id,
            "path": event.get("invocation", {}).get("invocationPath"),
        }

        # At the point in the code, the root log level seems to default to WARNING unless set to INFO, even with
        # the BasicConfig setting to INFO above.
        logging.root.setLevel(logging.INFO)

        logging_handler = fluenthandler.FluentHandler(
            "steamship.deployed_lambda",
            host=logging_host,
            port=logging_port,
            nanosecond_precision=True,
            msgpack_kwargs={"default": encode_exception},
        )

        # Without explicit instruction, the fluent handler defaults to UNSET. We want to make sure it is INFO.
        logging_handler.setLevel(logging.INFO)

        formatter = FluentRecordFormatter(custom_format)
        logging_handler.setFormatter(formatter)
        # The below should make it so calls to logging.info etc are also routed to the remote logger
        logging.root.addHandler(logging_handler)

    try:
        # Config will accept `workspace_id` as passed from the Steamship Engine, whereas the `Steamship`
        # class itself is limited to accepting `workspace` (`config.workspace_handle`) since that is the manner
        # of interaction ideal for developers.
        config = Configuration(**event.get("clientConfig", {}))
        client = Steamship(config=config, trust_workspace_config=True)
    except SteamshipError as se:
        logging.exception(se)
        return InvocableResponse.from_obj(se).dict(by_alias=True)
    except Exception as ex:
        logging.exception(ex)
        return InvocableResponse.error(
            code=HTTPStatus.INTERNAL_SERVER_ERROR,
            message="Plugin/App handler was unable to create Steamship client.",
            exception=ex,
        ).dict(by_alias=True)
    logging.info(f"Localstack hostname: {environ.get('LOCALSTACK_HOSTNAME')}.")
    response = internal_handler(event, client, invocation_context)

    result = response.dict(by_alias=True, exclude={"client"})
    # When created with data > 4MB, data is uploaded to a bucket.
    # This is a very ugly way to get the deep size of this object
    data = json.dumps(result.get("data", None)).encode("UTF-8")
    data_size = sys.getsizeof(data)
    logging.info(f"Response data size {data_size}")
    if data_size > 4e6 and invocation_context.invocable_type == "plugin":
        logging.info("Response data size >4MB, must upload to bucket")

        filepath = str(uuid.uuid4())
        signed_url = (
            client.get_workspace()
            .create_signed_url(
                SignedUrl.Request(
                    bucket=SignedUrl.Bucket.PLUGIN_DATA,
                    filepath=filepath,
                    operation=SignedUrl.Operation.WRITE,
                )
            )
            .signed_url
        )

        logging.info(f"Got signed url for writing: {signed_url}")

        upload_to_signed_url(signed_url, data)

        # Now remove raw data and replace with bucket
        del result["data"]
        result["dataBucket"] = SignedUrl.Bucket.PLUGIN_DATA.value
        result["dataFilepath"] = filepath

    if logging_handler is not None:
        logging_handler.close()

    return result


def create_handler(invocable_cls: Type[Invocable]):
    """Deprecated wrapper function for a Steamship invocable within an AWS Lambda function. Called by code within a
    plugin or package.
    """
    logging.warning(
        "Creating deprecated (unsafe imports) create_handler. This is no longer necessary. Please remove handler = create_handler(...) from your package or plugin."
    )

    def deprecated_handler(event, context=None):
        logging.error(
            "Calling deprecated (unsafe imports) create_handler. This indicates use of newer SDK against an older platform version."
        )

    return deprecated_handler


def safely_find_invocable_class() -> Type[Invocable]:
    """
    Safely find the invocable class within invocable code.
    """
    try:
        module = importlib.import_module("api")
        return get_class_from_module(module)
    except Exception as e:
        logging.exception(e)
        raise SteamshipError(
            message=f"There was an error loading the main file (it must be named api.py):\n{traceback.format_exc()}",
            error=e,
        )


def get_class_from_module(module) -> Type[Invocable]:
    invocable_classes = []
    for element in [getattr(module, x) for x in dir(module)]:
        if inspect.isclass(element):
            # Using names and not issubclass(element, Invocable) because latter was returning false?
            superclass_names = [c.__name__ for c in inspect.getmro(element)]
            if "Invocable" in superclass_names and element.__module__ == "api":
                invocable_classes.append(element)
    if len(invocable_classes) == 0:
        raise SteamshipError(
            message="Could not find package or plugin class in api.py. Define your package or plugin by subclassing from PluginService or PackageService."
        )
    if len(invocable_classes) > 1:
        raise SteamshipError(
            message=f"Found too many invocable classes {invocable_classes} in api.py. Only one is supported."
        )
    invocable_class = invocable_classes[0]
    logging.info(f"Safely loaded main class: {invocable_class.__name__}")
    return invocable_class


def create_safe_handler(known_invocable_for_testing: Type[Invocable] = None):
    if known_invocable_for_testing is not None:
        invocable_getter = lambda: known_invocable_for_testing  # noqa: E731
    else:
        invocable_getter = safely_find_invocable_class
    bound_internal_handler = lambda event, client, context: internal_handler(  # noqa: E731
        invocable_getter, event, client, context
    )
    return lambda event, context=None: handler(bound_internal_handler, event, context)


# safe_handler is the new handler entrypoint, allowing the import section of user-provided code to run in a
# context where we can trap errors.
safe_handler = create_safe_handler()



import pytest
from steamship_tests import TEST_ASSETS_PATH

from steamship import MimeTypes
from steamship.client import Steamship
from steamship.data.block import Block
from steamship.data.file import File
from steamship.data.tags.tag import Tag


@pytest.mark.usefixtures("client")
def test_query(client: Steamship):
    a = File.create(
        client=client,
        blocks=[
            Block(text="A", tags=[Tag(kind="BlockTag")]),
            Block(text="B"),
        ],
    )
    assert a.id is not None
    a = a.refresh()
    b = File.create(
        client=client,
        blocks=[
            Block(text="A"),
            Block(text="B", tags=[Tag(kind="Test")]),
        ],
        tags=[Tag(kind="FileTag")],
    )
    assert b.id is not None
    b = b.refresh()

    blocks = Block.query(client=client, tag_filter_query='blocktag and kind "BlockTag"').blocks
    assert len(blocks) == 1
    assert blocks[0].id == a.blocks[0].id

    blocks = Block.query(client=client, tag_filter_query='blocktag and kind "Test"').blocks
    assert len(blocks) == 1
    assert blocks[0].id == b.blocks[1].id

    a.delete()
    b.delete()


@pytest.mark.usefixtures("client")
def test_append_indices(client: Steamship):
    file = File.create(client, blocks=[Block(text="first")])
    assert len(file.blocks) == 1
    assert file.blocks[0].index_in_file == 0

    appended_block = Block.create(client, file_id=file.id, text="second")
    assert appended_block.index_in_file == 1

    file.refresh()
    assert len(file.blocks) == 2
    assert file.blocks[0].index_in_file == 0
    assert file.blocks[0].text == "first"
    assert file.blocks[1].index_in_file == 1
    assert file.blocks[1].text == "second"


@pytest.mark.usefixtures("client")
def test_append_block_url(client: Steamship):
    file = File.create(client, blocks=[])

    appended_block = Block.create(
        client, file_id=file.id, url="https://docs.steamship.com/_static/Steamship-symbol-dark.png"
    )

    assert appended_block.text == ""
    file.refresh()
    assert len(file.blocks) == 1
    assert file.blocks[0].mime_type == MimeTypes.PNG

    block_content = file.blocks[0].raw()
    assert len(block_content) == 20693
    assert not file.blocks[0].is_text()


@pytest.mark.usefixtures("client")
def test_text_mime_type(client: Steamship):
    file = File.create(client, blocks=[Block(text="first")])
    _ = Block.create(client, file_id=file.id, text="second")
    file.refresh()
    assert len(file.blocks) == 2
    for block in file.blocks:
        assert block.mime_type == MimeTypes.TXT
        assert block.is_text()


@pytest.mark.usefixtures("client")
def test_append_block_content_text(client: Steamship):
    file = File.create(client, blocks=[])

    appended_block = Block.create(client, file_id=file.id, content=bytes("This is a test", "utf-8"))

    assert appended_block.text == "This is a test"
    file.refresh()
    assert len(file.blocks) == 1
    assert file.blocks[0].mime_type == MimeTypes.TXT

    assert file.blocks[0].is_text()


@pytest.mark.usefixtures("client")
def test_append_block_content_image(client: Steamship):
    file = File.create(client, blocks=[])

    palm_tree_path = TEST_ASSETS_PATH / "palm_tree.png"
    with palm_tree_path.open("rb") as f:
        palm_bytes = f.read()

    appended_block = Block.create(
        client, file_id=file.id, content=palm_bytes, mime_type=MimeTypes.PNG
    )

    assert appended_block.text == ""
    file.refresh()
    assert len(file.blocks) == 1
    assert file.blocks[0].mime_type == MimeTypes.PNG

    assert not file.blocks[0].is_text()

    raw_content = file.blocks[0].raw()
    assert raw_content == palm_bytes


@pytest.mark.usefixtures("client")
def test_create_with_tags(client: Steamship):
    file = File.create(client, content="empty")
    new_block = Block.create(
        client,
        file_id=file.id,
        text="foo",
        tags=[Tag(kind="bar", name="foo"), Tag(kind="baz", name="foo")],
    )
    assert len(new_block.tags) == 2


@pytest.mark.usefixtures("client")
def test_append_is_present(client: Steamship):
    file = File.create(client, blocks=[Block(text="first")])
    assert len(file.blocks) == 1

    file.append_block(text="second")
    assert len(file.blocks) == 2

    file.refresh()
    assert len(file.blocks) == 2

    my_file = File.create(client, handle="my_file", content="")
    assert len(my_file.blocks) == 0
    my_file.append_block(text="first")
    assert len(my_file.blocks) == 1



from __future__ import annotations

from steamship.base.request import Request


class BlockifyRequest(Request):
    type: str = None
    plugin_instance: str = None
    id: str = None
    handle: str = None
    name: str = None



from contextlib import contextmanager
from typing import Any, Dict, Optional

from steamship import Configuration, Steamship, Task
from steamship.data.package.package_instance import PackageInstance
from steamship.data.plugin.plugin_instance import PluginInstance
from steamship.utils.url import Verb

TESTING_PROFILE = "test"


def get_steamship_client(fail_if_workspace_exists=False, **kwargs) -> Steamship:
    # Always use the `test` profile
    kwargs["profile"] = TESTING_PROFILE
    return Steamship(
        fail_if_workspace_exists=fail_if_workspace_exists, config=Configuration.parse_obj(kwargs)
    )


@contextmanager
def steamship_use(
    package_handle: str,
    instance_handle: str = None,
    config: Dict[str, Any] = None,
    version: str = None,
    fetch_if_exists: bool = True,
    delete_workspace: bool = True,
    **kwargs,
) -> PackageInstance:
    # Always use the `test` profile
    kwargs["profile"] = TESTING_PROFILE
    instance = Steamship.use(
        package_handle, instance_handle, config, version, fetch_if_exists, **kwargs
    )
    assert instance.client.config.workspace_id == instance.workspace_id
    yield instance
    # Clean up the workspace
    if delete_workspace:
        instance.client.get_workspace().delete()


@contextmanager
def steamship_use_as_if_deployed(klass):
    def add_method(klass, method, method_name=None):
        setattr(klass, method_name or method.__name__, method)

    def handle_kwargs(kwargs: Optional[dict] = None):
        if kwargs is not None and "wait_on_tasks" in kwargs:
            if kwargs["wait_on_tasks"] is not None:
                for task in kwargs["wait_on_tasks"]:
                    # It might not be of type Task if the invocation was something we've monkeypatched.
                    if type(task) == Task:
                        task.wait()
            kwargs.pop("wait_on_tasks")
        return kwargs

    def invoke(self, path: str, verb: Verb = Verb.POST, **kwargs):
        # Note: the correct impl would inspect the fn lookup for the fn with the right verb.
        path = path.rstrip("/").lstrip("/")
        fn = getattr(self, path)
        new_kwargs = handle_kwargs(kwargs)
        print(f"Patched invocation of self.invoke('{path}', {kwargs})")
        res = fn(**new_kwargs)
        if hasattr(res, "dict"):
            return res.dict()
        # TODO: Handle if they returned a InvocationResponse object
        return res

    def invoke_later(self, path: str, verb: Verb = Verb.POST, **kwargs):
        # Note: the correct impl would inspect the fn lookup for the fn with the right verb.
        path = path.rstrip("/").lstrip("/")
        fn = getattr(self, path)
        new_kwargs = handle_kwargs(kwargs)
        invoke_later_args = new_kwargs.get("arguments", {})  # Specific to invoke_later
        print(f"Patched invocation of self.invoke_later('{path}', {kwargs})")
        return fn(**invoke_later_args)

    client = Steamship()
    add_method(klass, invoke)
    add_method(klass, invoke_later)
    obj = klass(client=client)

    # TODO: Clean up workspace when done
    yield obj
    return obj


@contextmanager
def steamship_use_plugin(
    plugin_handle: str,
    instance_handle: str = None,
    config: Dict[str, Any] = None,
    version: str = None,
    fetch_if_exists: bool = True,
    delete_workspace: bool = True,
    **kwargs,
) -> PluginInstance:
    # Always use the `test` profile
    kwargs["profile"] = TESTING_PROFILE
    instance = Steamship.use_plugin(
        plugin_handle, instance_handle, config, version, fetch_if_exists, **kwargs
    )
    assert instance.client.config.workspace_id == instance.workspace_id
    yield instance
    # Clean up the workspace
    if delete_workspace:
        instance.client.get_workspace().delete()


@contextmanager
def steamship_use_skill(
    skill: str,
    provider: Optional[str] = None,
    instance_handle: str = None,
    fetch_if_exists: bool = True,
    delete_workspace: bool = True,
) -> PluginInstance:
    # TODO (enias): Always use the same steamship instance for testing
    instance = Steamship(profile=TESTING_PROFILE).use_skill(
        skill=skill,
        provider=provider,
        instance_handle=instance_handle,
        fetch_if_exists=fetch_if_exists,
    )
    assert instance.client.config.workspace_id == instance.workspace_id
    yield instance
    # Clean up the workspace
    if delete_workspace:
        instance.client.get_workspace().delete()


def register_plugin_instance_subclass(plugin_handle: str, plugin_class: PluginInstance):
    """Register that subclass `plugin_class` should be used for instances of plugin handle `plugin_handle`."""
    Steamship._PLUGIN_INSTANCE_SUBCLASS_OVERRIDES[plugin_handle] = plugin_class



from __future__ import annotations

from typing import Any, Dict

from steamship.base import Configuration
from steamship.base.model import CamelModel


class Invocation(CamelModel):
    http_verb: str = None
    invocation_path: str = None  # e.g. /hello/there
    arguments: Dict[str, Any] = None
    config: Dict[str, Any] = None


class LoggingConfig(CamelModel):
    logging_host: str = None
    logging_port: str = None


class InvocationContext(CamelModel):
    tenant_id: str = None
    user_id: str = None
    workspace_id: str = None
    invocable_handle: str = None
    invocable_version_handle: str = None
    invocable_instance_handle: str = None
    invocable_type: str = None
    invocable_owner_id: str = None


class InvocableRequest(CamelModel):
    """A request as the Steamship Hosting Framework receives it from the Engine.

    This class is different from the other `Request` class:
     * `steamship.base.request` represents a request from the Steamship Client
     * this class represents a request from the Steamship Engine to a Steamship-hosted App/Plugin

    It contains both a package/plugin invocation and also the client configuration in which that invocation
    is intended to execute.
    """

    client_config: Configuration = None
    invocation: Invocation = None
    logging_config: LoggingConfig = None
    invocation_context: InvocationContext = None



from .configuration import Configuration
from .environments import RuntimeEnvironments, check_environment
from .error import SteamshipError
from .mime_types import MimeTypes
from .tasks import Task, TaskState

__all__ = [
    "Configuration",
    "SteamshipError",
    "Task",
    "TaskState",
    "MimeTypes",
    "RuntimeEnvironments",
    "check_environment",
]



import logging

import pytest
from assets.plugins.taggers.plugin_third_party_trainable_tagger import MockClient
from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin

from steamship import Steamship
from steamship.base import TaskState
from steamship.data.plugin import HostingType
from steamship.data.plugin.plugin_instance import PluginInstance
from steamship.plugin.inputs.export_plugin_input import ExportPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput

EXPORTER_HANDLE = "signed-url-exporter"


@pytest.mark.usefixtures("client")
def test_e2e_third_party_trainable_tagger_lambda_training(client: Steamship):
    exporter_plugin_r = PluginInstance.create(
        client=client,
        handle=EXPORTER_HANDLE,
        plugin_handle=EXPORTER_HANDLE,
        fetch_if_exists=True,  # Don't care if it already exists
    )
    assert exporter_plugin_r is not None
    exporter_plugin = exporter_plugin_r
    assert exporter_plugin.handle is not None

    third_party_trainable_tagger_path = (
        PLUGINS_PATH / "taggers" / "plugin_third_party_trainable_tagger.py"
    )

    # Note that we're going to do the below training on ZERO data for simplicity.
    # The particular test model doesn't actually incorporate any data given to it at training time, so
    # it would just slow the test down to create, blockify, and export a training corpus.

    with deploy_plugin(
        client, third_party_trainable_tagger_path, "tagger", training_platform=HostingType.LAMBDA
    ) as (tagger, tagger_version, tagger_instance):
        # Now train the plugin
        training_request = TrainingParameterPluginInput(
            plugin_instance=tagger_instance.handle,
            export_plugin_input=ExportPluginInput(
                plugin_instance=exporter_plugin.handle, type="file", query="all"
            ),
        )
        train_result = tagger_instance.train(training_request)
        train_result.wait()
        assert train_result.state is not TaskState.failed
        assert train_result.remote_status_input is not None
        assert train_result.remote_status_input["num_checkins"] == 2

        assert train_result.output is not None

        logging.info("Waiting 15 seconds for instance to deploy.")
        import time

        time.sleep(15)

        # Now we'll attempt to USE this plugin. This plugin's behavior is to simply tag every block with
        # the parameters `MockClient.LABELS`

        # First we'll create a file
        test_doc = "Hi there"
        res = tagger_instance.tag(doc=test_doc)
        res.wait()
        assert res.output is not None
        assert res.output.file is not None
        assert not res.output.file.tags
        assert res.output.file.blocks is not None
        assert len(res.output.file.blocks) > 0
        for block in res.output.file.blocks:
            assert block.tags is not None
            assert sorted([tag.kind for tag in block.tags]) == sorted(MockClient.LABELS)



import os
import webbrowser
from unittest import mock

import pytest
from pydantic import ValidationError

from steamship import Configuration, SteamshipError
from steamship.base.configuration import DEFAULT_API_BASE, DEFAULT_APP_BASE, DEFAULT_WEB_BASE

TEST_WEB_BASE = "https://app.test.com/"
TEST_APP_BASE = "https://test.run/"
TEST_API_BASE = "https://api.test.com/api/v1/"

empty_base_uris = [
    (TEST_WEB_BASE, TEST_APP_BASE, TEST_API_BASE),
    (None, TEST_APP_BASE, TEST_API_BASE),
    (TEST_WEB_BASE, None, TEST_API_BASE),
    (TEST_WEB_BASE, TEST_APP_BASE, None),
]


@pytest.mark.parametrize(("web_base", "app_base", "api_base"), empty_base_uris)
def test_base_uris(web_base: str, app_base: str, api_base: str) -> None:
    configuration = Configuration(api_base=api_base, web_base=web_base, app_base=app_base)
    assert str(configuration.web_base) == web_base or DEFAULT_WEB_BASE
    assert str(configuration.app_base) == app_base or DEFAULT_APP_BASE
    assert str(configuration.api_base) == api_base or DEFAULT_API_BASE


def test_incorrectly_formatted_base_uris() -> None:
    configuration = Configuration(
        api_base=TEST_API_BASE[:-1], web_base=TEST_WEB_BASE[:-1], app_base=TEST_APP_BASE[:-1]
    )
    assert str(configuration.web_base) == TEST_WEB_BASE
    assert str(configuration.app_base) == TEST_APP_BASE
    assert str(configuration.api_base) == TEST_API_BASE


@pytest.mark.parametrize("base_uri", ["", "test", "ftp://test.com"])
def test_invalid_base_uris(base_uri: str) -> None:
    with pytest.raises(ValidationError):
        Configuration(api_base=base_uri, web_base=base_uri, app_base=base_uri)


def test_empty_base_uris() -> None:
    configuration = Configuration()
    assert configuration.web_base is not None
    assert configuration.app_base is not None
    assert configuration.api_base is not None


@mock.patch.dict(os.environ, {"STEAMSHIP_API_KEY": ""})
def test_empty_api_key() -> None:

    # Only run this test on non-local (i.e., CI/CD) otherwise will attempt browser login
    try:
        _ = webbrowser.get()
    except webbrowser.Error:
        with pytest.raises(SteamshipError):
            # Note: We're referencing a non existing profile to make sure the api key is not loaded from the default profile in steamship.json
            Configuration(api_key=None, profile="non-existing-profile")



import importlib.machinery as machinery
import os
import sys
import traceback
import zipfile
from abc import ABC, abstractmethod
from pathlib import Path

import click
from semver import VersionInfo

from steamship import Package, PackageVersion, PluginVersion, Steamship, SteamshipError
from steamship.cli.manifest_init_wizard import validate_handle, validate_version_handle
from steamship.cli.ship_spinner import ship_spinner
from steamship.data import Plugin
from steamship.data.manifest import Manifest
from steamship.data.user import User
from steamship.invocable.lambda_handler import get_class_from_module

DEFAULT_BUILD_IGNORE = [
    "build",
    ".git",
    ".venv",
    ".ipynb_checkpoints",
    ".DS_Store",
    "venv",
    "tests",
    "examples",
    ".idea",
    "__pycache__",
]


def update_config_template(manifest: Manifest):

    path = Path("src/api.py")
    if not path.exists():
        path = Path("api.py")
        if not path.exists():
            raise SteamshipError("Could not find api.py either in root directory or in src.")

    api_module = None
    try:
        sys.path.append(str(path.parent.absolute()))

        # load the API module to allow config inspection / generation
        api_module = machinery.SourceFileLoader("api", str(path)).load_module()
    except Exception:
        click.secho(
            "An error occurred while loading your api.py to check configuration parameters. Full stack trace below.",
            fg="red",
        )
        traceback.print_exc()
        click.get_current_context().abort()

    invocable_type = get_class_from_module(api_module)

    config_parameters = invocable_type.config_cls().get_config_parameters()

    if manifest.configTemplate != config_parameters:
        if len(config_parameters) > 0:
            click.secho("Config parameters changed; updating steamship.json.", fg="cyan")
            for param_name, param in config_parameters.items():
                click.echo(f"{param_name}:")
                click.echo(f"\tType: {param.type}")
                click.echo(f"\tDefault: {param.default}")
                click.echo(f"\tDescription: {param.description}")
        else:
            click.secho("Config parameters removed; updating steamship.json.", fg="cyan")

    manifest.configTemplate = config_parameters
    manifest.save()


def get_archive_path(manifest: Manifest) -> Path:
    return Path(".") / "build" / "archives" / f"{manifest.handle}_v{manifest.version}.zip"


def bundle_deployable(manifest: Manifest):
    archive_path = get_archive_path(manifest)
    archive_path.parent.mkdir(parents=True, exist_ok=True)
    excludes = DEFAULT_BUILD_IGNORE + manifest.build_config.get("ignore", [])

    archive_path.unlink(missing_ok=True)

    # This zipfile packaging is modeled after the typescript CLI.
    # Items in non-excluded root folders are moved to the top-level.

    with zipfile.ZipFile(
        file=archive_path, mode="a", compression=zipfile.ZIP_DEFLATED, allowZip64=False
    ) as zip_file:

        root = Path(".")
        for file_path in root.iterdir():
            if file_path.name not in excludes:
                if file_path.is_dir():
                    for directory, _, files in os.walk(file_path):
                        subdirectory_path = Path(directory)
                        if Path(directory).name not in excludes:
                            for file in files:
                                pypi_file = subdirectory_path / file
                                relative_to = pypi_file.relative_to(file_path)
                                zip_file.write(pypi_file, relative_to)

                else:
                    zip_file.write(file_path)


class DeployableDeployer(ABC):
    @abstractmethod
    def _create_version(self, client: Steamship, manifest: Manifest, thing_id: str):
        pass

    @abstractmethod
    def create_object(self, client: Steamship, manifest: Manifest):
        pass

    @abstractmethod
    def update_object(self, deployable, client: Steamship, manifest: Manifest):
        pass

    @abstractmethod
    def deployable_type(self):
        pass

    def create_or_fetch_deployable(self, client: Steamship, user: User, manifest: Manifest):
        if not manifest.handle or len(manifest.handle) == 0:
            self.ask_for_new_handle(manifest, was_missing=True)

        deployable = None
        while deployable is None:
            click.echo(
                f"Creating / fetching {self.deployable_type()} with handle [{manifest.handle}]... ",
                nl=False,
            )
            try:
                deployable = self.create_object(client, manifest)
                if deployable.user_id != user.id:
                    self.ask_for_new_handle(manifest)
                    deployable = None
            except SteamshipError as e:
                if e.message == "Something went wrong.":
                    self.ask_for_new_handle(manifest)
                else:
                    click.secho(
                        f"Unable to create / fetch {self.deployable_type()}. Server returned message: {e.message}"
                    )
                    click.get_current_context().abort()

        self.update_object(deployable, client, manifest)

        click.echo("Done.")
        return deployable

    def ask_for_new_handle(self, manifest: Manifest, was_missing: bool = False):
        if not was_missing:
            try_again = click.confirm(
                click.style(
                    f"\nIt looks like that handle [{manifest.handle}] is already in use. Would you like to change the handle and try again?",
                    fg="yellow",
                ),
                default=True,
            )
            if not try_again:
                click.get_current_context().abort()

        new_handle = click.prompt(
            f"What handle would you like to use for your {self.deployable_type()}? Valid characters are a-z and -",
            value_proc=validate_handle,
        )
        manifest.handle = new_handle
        manifest.save()

    def create_version(self, client: Steamship, manifest: Manifest, thing_id: str):
        version = None

        if not manifest.version or len(manifest.version) == 0:
            self.ask_for_new_version_handle(manifest, was_missing=True)

        while version is None:
            click.echo(f"Deploying version {manifest.version} of [{manifest.handle}]... ", nl=False)
            try:
                with ship_spinner():
                    version = self._create_version(client, manifest, thing_id)
            except SteamshipError as e:
                if "The object you are trying to create already exists." in e.message:
                    self.ask_for_new_version_handle(manifest)
                else:
                    click.secho(f"\nUnable to deploy {self.deployable_type()} version.", fg="red")
                    click.secho(f"Server returned message: {e.message}")
                    if "ModuleNotFoundError" in e.message:
                        click.secho(
                            "It looks like you may be missing a dependency in your requirements.txt.",
                            fg="yellow",
                        )
                    click.get_current_context().abort()
        click.echo("\nDone. 🚢")

    def ask_for_new_version_handle(self, manifest: Manifest, was_missing: bool = False):
        if not was_missing:
            try_again = click.confirm(
                click.style(
                    f"\nIt looks like that version [{manifest.version}] has already been deployed. Would you like to change the version handle and try again?",
                    fg="yellow",
                ),
                default=True,
            )
            if not try_again:
                click.get_current_context().abort()

        default_new = "1.0.0"
        try:
            default_new = str(VersionInfo.parse(manifest.version).bump_prerelease())
        except ValueError:
            pass
        old_archive_path = get_archive_path(manifest)
        new_version_handle = click.prompt(
            "What should the new version be? Valid characters are a-z, 0-9, . and -",
            value_proc=validate_version_handle,
            default=default_new,
        )
        manifest.version = new_version_handle
        manifest.save()
        new_archive_path = get_archive_path(manifest)
        old_archive_path.rename(new_archive_path)


class PackageDeployer(DeployableDeployer):
    def _create_version(self, client: Steamship, manifest: Manifest, thing_id: str):
        return PackageVersion.create(
            client=client,
            config_template=manifest.config_template_as_dict(),
            handle=manifest.version,
            filename=f"build/archives/{manifest.handle}_v{manifest.version}.zip",
            package_id=thing_id,
        )

    def create_object(self, client: Steamship, manifest: Manifest):
        return Package.create(
            client,
            handle=manifest.handle,
            profile=manifest,
            is_public=manifest.public,
            fetch_if_exists=True,
        )

    def update_object(self, deployable, client: Steamship, manifest: Manifest):
        deployable.profile = manifest

        package = deployable.update(client)
        return package

    def deployable_type(self):
        return "package"


class PluginDeployer(DeployableDeployer):
    def _create_version(self, client: Steamship, manifest: Manifest, thing_id: str):
        return PluginVersion.create(
            client=client,
            config_template=manifest.config_template_as_dict(),
            handle=manifest.version,
            filename=f"build/archives/{manifest.handle}_v{manifest.version}.zip",
            plugin_id=thing_id,
        )

    def create_object(self, client: Steamship, manifest: Manifest):
        return Plugin.create(
            client,
            description=manifest.description,
            is_public=manifest.public,
            transport=manifest.plugin.transport,
            type_=manifest.plugin.type,
            handle=manifest.handle,
            fetch_if_exists=True,
        )

    def update_object(self, deployable, client: Steamship, manifest: Manifest):
        deployable.profile = manifest

        plugin = deployable.update(client)
        return plugin

    def deployable_type(self):
        return "plugin"



from __future__ import annotations

from enum import Enum
from typing import Any, List, Optional, Type, Union

import requests
from pydantic import BaseModel, Field

from steamship import MimeTypes, SteamshipError
from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import DeleteRequest, IdentifierRequest, Request
from steamship.base.response import Response
from steamship.data.tags.tag import Tag


class BlockQueryRequest(Request):
    tag_filter_query: str


class BlockUploadType(str, Enum):
    FILE = "file"  # A file uploaded as bytes or a string
    BLOCKS = "blocks"  # Blocks are sent to create a file
    URL = "url"  # content will be fetched from a URL
    NONE = "none"  # No upload; plain text only


class Block(CamelModel):
    """A Block is a chunk of content within a File. It can be plain text content, image content,
    video content, etc. If the content is not text, the text value may be the empty string
    for backwards compatibility.
    """

    client: Client = Field(None, exclude=True)
    id: str = None
    file_id: str = None
    text: str = None
    tags: Optional[List[Tag]] = []
    index_in_file: Optional[int] = Field(alias="index")
    mime_type: Optional[MimeTypes]
    url: Optional[
        str
    ] = None  # Only for creation of blocks; used to fetch content from a public URL.
    content_url: Optional[
        str
    ] = None  # For overriding the URL of the raw data for ephemeral blocks. Setting this will have no effect
    upload_type: Optional[
        BlockUploadType
    ] = None  # for returning Blocks as the result of a generate request

    class ListRequest(Request):
        file_id: str = None

    class ListResponse(Response):
        blocks: List[Block] = []

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["block"] if "block" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def get(
        client: Client,
        _id: str = None,
    ) -> Block:
        return client.post(
            "block/get",
            IdentifierRequest(id=_id),
            expect=Block,
        )

    @staticmethod
    def create(
        client: Client,
        file_id: str,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """
        Create a new Block within a File specified by file_id.

        You can create a Block in several ways:
        - Providing raw text as the text parameter;
        - Providing the content of the block as string or bytes;
        - Providing a publicly accessible URL where the content is stored.

        """

        if content is not None and url is not None:
            raise SteamshipError("May provide content or URL, but not both when creating a Block")

        if content is not None:
            upload_type = BlockUploadType.FILE
        elif url is not None:
            upload_type = BlockUploadType.URL
        else:
            upload_type = BlockUploadType.NONE

        req = {
            "fileId": file_id,
            "text": text,
            "tags": [t.dict(by_alias=True) for t in tags] if tags else [],
            "url": url,
            "mimeType": mime_type,
            "uploadType": upload_type,
        }

        file_data = (
            ("file-part", content, "multipart/form-data")
            if upload_type == BlockUploadType.FILE
            else None
        )

        return client.post(
            "block/create",
            req,
            expect=Block,
            file=file_data,
        )

    def delete(self) -> Block:
        return self.client.post(
            "block/delete",
            DeleteRequest(id=self.id),
            expect=Tag,
        )

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
    ) -> BlockQueryResponse:
        req = BlockQueryRequest(tag_filter_query=tag_filter_query)
        res = client.post(
            "block/query",
            payload=req,
            expect=BlockQueryResponse,
        )
        return res

    def index(self, embedding_plugin_instance: Any = None):
        """Index this block."""
        tags = [
            Tag(
                text=self.text,
                file_id=self.file_id,
                block_id=self.id,
                kind="block",
                start_idx=0,
                end_idx=len(self.text),
            )
        ]
        return embedding_plugin_instance.insert(tags)

    def raw(self):
        if self.content_url is not None:
            return requests.get(self.content_url).content
        else:
            return self.client.post(
                "block/raw",
                payload={
                    "id": self.id,
                },
                raw_response=True,
            )

    def is_text(self):
        """Return whether this is a text Block."""
        return self.mime_type == MimeTypes.TXT


class BlockQueryResponse(Response):
    blocks: List[Block]


Block.ListResponse.update_forward_refs()
Block.update_forward_refs()



# TODO: It should fail if the docs field is empty.
# TODO: It should fail if the file hasn't been converted.
from steamship_tests.utils.fixtures import get_steamship_client
from steamship_tests.utils.random import random_index

from steamship import DocTag, File, MimeTypes, PluginInstance, Tag

_TEST_EMBEDDER = "test-embedder"

T = "A nice poem"
P1_1 = "Roses are red."
P1_2 = "Violets are blue."
P2_1 = "Sugar is sweet."
P2_2 = "I love you."
T2 = "A flavorful story"
P3_1 = "Cake is made of flour."
P3_2 = "Cake tastes good with milk."
P4_1 = "Cake comes in chocolate and vanilla flavors."
P4_2 = "Cake can be cut into mAny pieces and shared."


def test_file_parse():
    steamship = get_steamship_client()
    content1 = f"# {T}\n\n{P1_1} {P1_2}\n\n{P2_1} {P2_2}"
    content2 = f"# {T2}\n\n{P3_1} {P3_2}\n\n{P4_1} {P4_2}"
    content = f"{content1}\n\n{content2}"

    file = File.create(
        steamship,
        content=content,
        mime_type=MimeTypes.MKD,
    )
    assert file.id is not None
    assert file.mime_type == MimeTypes.MKD

    blockify_resp = file.blockify(plugin_instance="markdown-blockifier-default-1.0")
    blockify_resp.wait()

    # Now we parse
    parser = PluginInstance.create(steamship, plugin_handle="test-tagger")
    parse_resp = file.tag(plugin_instance=parser.handle)
    parse_resp.wait()

    # Now the sentences should be parsed!
    q2 = file.refresh()
    assert len(q2.blocks) == 6

    # Now we add the file to the index
    with random_index(steamship, plugin_instance=_TEST_EMBEDDER) as index:
        file.index(index)
        res = index.search("What color are roses?")
        res.wait()
        items = res.output.items
        assert len(items) == 1
        # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
        assert items[0].tag.text == " ".join([P1_1, P1_2])
        assert items[0].tag.file_id == file.id
        assert items[0].tag.block_id is not None

    file.delete()


def test_file_index():
    # TODO(ted) - Refactor this to the new index plugin style on a later pass.
    steamship = get_steamship_client()
    t = "A nice poem"
    p1_1 = "Roses are red."
    p1_2 = "Violets are blue."
    p2_1 = "Sugar is sweet."
    p2_2 = "I love you."
    t2 = "A flavorful story"
    p3_1 = "Cake is made of flour."
    p3_2 = "Cake tastes good with milk."
    p4_1 = "Cake comes in chocolate and vanilla flavors."
    p4_2 = "Cake can be cut into mAny pieces and shared."

    content1 = f"# {t}\n\n{p1_1} {p1_2}\n\n{p2_1} {p2_2}"
    content2 = f"# {t2}\n\n{p3_1} {p3_2}\n\n{p4_1} {p4_2}"
    content = f"{content1}\n\n{content2}"

    file = File.create(
        steamship,
        content=content,
        mime_type=MimeTypes.MKD,
    )
    assert file.id is not None
    assert file.mime_type == MimeTypes.MKD

    blockify_resp = file.blockify(plugin_instance="markdown-blockifier-default-1.0")
    blockify_resp.wait()

    # Now we parse
    parser = PluginInstance.create(steamship, plugin_handle="test-tagger")
    parse_resp = file.tag(plugin_instance=parser.handle)
    parse_resp.wait()

    # Now the sentences should be parsed!
    q2 = file.refresh()
    assert len(q2.blocks) == 6

    with random_index(steamship, plugin_instance=_TEST_EMBEDDER) as index:
        file.index(index)

        res = index.search("What color are roses?")
        res.wait()
        items = res.output.items
        assert len(items) == 1
        # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
        assert items[0].tag.text == " ".join([p1_1, p1_2])

        res = index.search("What flavors does cake come in?")
        res.wait()
        items = res.output.items
        assert len(items) == 1
        # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
        assert items[0].tag.text == " ".join([p4_1, p4_2])

    file.delete()


def test_file_embed_lookup():
    steamship = get_steamship_client()

    content_a = "Ted likes to run."
    content_b = "Grace likes to bike."

    file_1 = File.create(
        steamship,
        content=content_a,
        mime_type=MimeTypes.MKD,
    )

    blockify_task = file_1.blockify(plugin_instance="markdown-blockifier-default-1.0")
    blockify_task.wait()

    parser = PluginInstance.create(steamship, plugin_handle="test-tagger")
    parse_res = file_1.tag(plugin_instance=parser.handle)
    parse_res.wait()

    file_2 = File.create(client=steamship, content=content_b, mime_type=MimeTypes.MKD)
    blockify_task = file_2.blockify(plugin_instance="markdown-blockifier-default-1.0")
    blockify_task.wait()

    parser = PluginInstance.create(steamship, plugin_handle="test-tagger")
    parse_res = file_2.tag(plugin_instance=parser.handle)
    parse_res.wait()

    # Now we add the file to the index
    with random_index(steamship, _TEST_EMBEDDER) as index:
        # Insert the tags from file

        # TODO(dave, ted) - Note for future.
        # This is a really interesting place for us to experiment with tag.text, query results, etc.
        #
        # It would be interesting if we could do something like:
        #
        #   File.index(index, tag_query='name "sentence"')
        #
        # Which I think might require something like:
        #   - Query to return Tag.text filled out
        #   - Query system to be able to auto-prepend "file_id {file.id}" in a way that "just works"
        #     (e.g. would this be a string operation in string query-rewriting space, or a scoping within which
        #           the un-modified query was executed)
        #
        # At the moment, I think the below is what it takes to get sentence tags from a file into an index.
        res = Tag.query(
            steamship, f"""blocktag and file_id "{file_1.id}" and name "{DocTag.SENTENCE}" """
        )

        # Fill in the tag.text field
        file_1 = file_1.refresh()
        file_1_tag = None
        for tag in res.tags:
            for block in file_1.blocks:
                if tag.block_id == block.id:
                    tag.text = block.text[tag.start_idx : tag.end_idx]
                    file_1_tag = tag

        index.insert(res.tags)

        res = Tag.query(
            steamship, f"""blocktag and file_id "{file_2.id}" and name "{DocTag.SENTENCE}" """
        )

        # Fill in the tag.text field
        file_2 = file_2.refresh()
        file_2_tag = None
        for tag in res.tags:
            for block in file_2.blocks:
                if tag.block_id == block.id:
                    tag.text = block.text[tag.start_idx : tag.end_idx]
                    file_2_tag = tag

        index.insert(res.tags)

        res = index.search("What does Ted like to do?")
        res.wait()
        items = res.output.items
        assert len(items) == 1
        assert items[0].tag.text == file_1_tag.text

        res = index.search("What does Grace like to do?")
        res.wait()
        items = res.output.items
        assert len(items) == 1
        assert items[0].tag.text == file_2_tag.text



#
# This is the CLIENT-side abstraction for an invocable.
#
# If you are implementing a package, see: steamship.invocable.server.App
#

from __future__ import annotations

from typing import Any, Optional, Type

from pydantic import BaseModel, Field

from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import CreateRequest, GetRequest, UpdateRequest
from steamship.data.manifest import Manifest


class PackageCreateRequest(CreateRequest):
    is_public: bool = False
    fetch_if_exists = False
    profile: Optional[Manifest] = None


class PackageUpdateRequest(UpdateRequest):
    id: Optional[str] = None
    handle: Optional[str] = None
    description: Optional[str] = None
    profile: Optional[Manifest] = None
    readme: Optional[str] = None


class Package(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    handle: str = None
    user_id: str = None
    profile: Optional[Manifest] = None
    description: Optional[str] = None
    readme: Optional[str] = None
    is_public: bool = False

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["package"] if "package" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        handle: str = None,
        profile: Manifest = None,
        is_public=False,
        fetch_if_exists=False,
    ) -> Package:
        req = PackageCreateRequest(
            handle=handle, profile=profile, is_public=is_public, fetch_if_exists=fetch_if_exists
        )
        return client.post("package/create", payload=req, expect=Package)

    @staticmethod
    def get(client: Client, handle: str) -> Package:
        return client.post("package/get", GetRequest(handle=handle), expect=Package)

    def update(self, client: Client) -> Package:
        return client.post(
            "package/update",
            PackageUpdateRequest(
                id=self.id, description=self.description, profile=self.profile, readme=self.readme
            ),
            expect=Package,
        )



from abc import abstractmethod
from typing import List, Optional

from steamship import Block, File, SteamshipError, Tag, Task, TaskState
from steamship.invocable import InvocableResponse
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.blockifier.blockifier import Blockifier
from steamship.plugin.inputs.raw_data_plugin_input import RawDataPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput

TRANSCRIPT_ID = "transcript_id"


class Transcriber(Blockifier):
    @abstractmethod
    def start_transcription(self, audio_file: PluginRequest[RawDataPluginInput]) -> str:
        """Start a transcription job and return an id to identify the transcription."""
        raise NotImplementedError()

    @abstractmethod
    def get_transcript(self, transcript_id: str) -> (Optional[str], Optional[List[Tag]]):
        """Method to retrieve the transcript and optional Tags. If the transcription is not ready, return None"""
        raise NotImplementedError()

    def _get_transcript(self, transcript_id: str) -> InvocableResponse:
        """Retrieve the transcript using the transcript_id."""
        transcript, tags = self.get_transcript(transcript_id)
        if transcript is None and tags is None:
            return InvocableResponse(
                status=Task(
                    state=TaskState.running,
                    remote_status_message="Transcription is ongoing.",
                    remote_status_input={"transcript_id": transcript_id},
                )
            )
        else:
            return InvocableResponse(
                data=BlockAndTagPluginOutput(
                    file=File(
                        blocks=[
                            Block(
                                text=transcript,
                                tags=tags,
                            )
                        ]
                    )
                )
            )

    def run(
        self, request: PluginRequest[RawDataPluginInput]
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        if request.is_status_check:
            if TRANSCRIPT_ID not in request.status.remote_status_input:
                raise SteamshipError(message="Status check requests need to provide a valid job id")
            transcript_id = request.status.remote_status_input[TRANSCRIPT_ID]
            return self._get_transcript(transcript_id)

        else:
            transcript_id = self.start_transcription(audio_file=request.data.data)
            return self._get_transcript(transcript_id)



"""Please see https://docs.steamship.com/ for information about building a Steamship Package"""
import inspect
import logging
import pathlib
import time
from abc import ABC
from collections import defaultdict
from functools import wraps
from http import HTTPStatus
from typing import Any, Dict, Optional, Type, Union

import toml

from steamship.base.package_spec import MethodSpec, PackageSpec
from steamship.client.steamship import Steamship
from steamship.invocable import Config
from steamship.invocable.config import ConfigParameter
from steamship.invocable.invocable_request import InvocableRequest, InvocationContext
from steamship.invocable.invocable_response import InvocableResponse
from steamship.utils.url import Verb


def make_registering_decorator(decorator):
    """
    Returns a copy of foreignDecorator, which is identical in every
    way(*), except also appends a .decorator property to the callable it
    spits out.

    (*)We can be somewhat "hygienic", but newDecorator still isn't signature-preserving,
    i.e. you will not be able to get a runtime list of parameters.
    For that, you need hackish libraries...but in this case, the only argument is func, so it's not a big issue
    """

    def new_decorator(func):
        # Call to newDecorator(method)
        # Exactly like old decorator, but output keeps track of what decorated it
        output = decorator(
            func
        )  # apply foreignDecorator, like call to foreignDecorator(method) would have done
        output.decorator = new_decorator  # keep track of decorator
        # R.original = func         # might as well keep track of everything!
        return output

    new_decorator.__name__ = decorator.__name__
    new_decorator.__doc__ = decorator.__doc__
    new_decorator.__is_endpoint__ = True
    return new_decorator


# https://stackoverflow.com/questions/2366713/can-a-decorator-of-an-instance-method-access-the-class
# noinspection PyUnusedLocal
def endpoint(verb: str = None, path: str = None, **kwargs):
    """By using ``kwargs`` we can tag the function with Any parameters."""  # noqa: RST210

    def decorator(function):
        # This is used in conjunction with the __init_subclass__ code!
        # Otherwise the __name__ won't be correct in maybeDecorated.__name__!
        # noinspection PyShadowingNames
        @wraps(function)
        def wrap(self, *args, **kwargs):
            return function(self, *args, **kwargs)

        # Build a dictionary of String->Primitive Types to pass back with endpoint
        # This enables the Engine to add support for features like public=True, etc, without the Client changing.
        config: Dict[str, Union[str, bool, int, float]] = {}
        for key, val in kwargs.items():
            if isinstance(val, (str, bool, int, float)):
                config[key] = val

        wrap.__path__ = path
        wrap.__verb__ = verb
        wrap.__endpoint_config__ = config

        return wrap

    decorator = make_registering_decorator(decorator)
    return decorator


def get(path: str, **kwargs):
    return endpoint(verb=Verb.GET, path=path, **kwargs)


def post(path: str, **kwargs):
    return endpoint(verb=Verb.POST, path=path, **kwargs)


class Invocable(ABC):
    """A Steamship microservice.

    This model.py class:

      1. Provide a pre-authenticated instance of the Steamship client
      2. Provides a Lambda handler that routes to registered functions
      3. Provides useful methods connecting functions to the router.
    """

    _method_mappings = defaultdict(dict)
    _package_spec: PackageSpec
    config: Config
    context: InvocationContext

    def __init__(
        self,
        client: Steamship = None,
        config: Dict[str, Any] = None,
        context: InvocationContext = None,
    ):
        self.context = context

        try:
            secret_kwargs = toml.load(".steamship/secrets.toml")
        except FileNotFoundError:  # Support local secret loading
            try:
                local_secrets_file = (
                    pathlib.Path(inspect.getfile(type(self))).parent / ".steamship" / "secrets.toml"
                )
                secret_kwargs = toml.load(str(local_secrets_file))
            except (TypeError, FileNotFoundError):
                secret_kwargs = {}

        # The configuration for the Invocable is the union of:
        #
        # 1) The `secret_kwargs` dict, read in from .steamship/secrets.toml, if it exists, and
        # 2) The `config` dict, provided upon instantiation.
        #
        # When invoked from within Steamship, the `config` dict is frozen, at the instance level, upon instance
        # creation. All subsequent method invocations reuse that frozen config.
        config = {
            **secret_kwargs,
            **{k: v for k, v in (config or {}).items() if v != ""},
        }

        # Finally, we set the config object to an instance of the class returned by `self.config_cls`
        if config:
            self.config = self.config_cls()(**config)
        else:
            self.config = self.config_cls()()

        self.client = client

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        start_time = time.time()
        cls._package_spec = PackageSpec(name=cls.__name__, doc=cls.__doc__, methods=[])
        cls._method_mappings = defaultdict(dict)
        base_fn_list = [
            may_be_decorated
            for base_cls in cls.__bases__
            for may_be_decorated in base_cls.__dict__.values()
        ]
        for attribute in base_fn_list + list(cls.__dict__.values()):
            decorator = getattr(attribute, "decorator", None)
            if decorator:
                if getattr(decorator, "__is_endpoint__", False):
                    path = getattr(attribute, "__path__", None)
                    verb = getattr(attribute, "__verb__", None)
                    config = getattr(attribute, "__endpoint_config__", {})
                    method_spec = cls._register_mapping(
                        name=attribute.__name__, verb=verb, path=path, config=config
                    )
                    cls._package_spec.methods.append(method_spec)

        # Add the HTTP GET /__dir__ method which returns a serialization of the PackageSpec.
        # Wired up to both GET and POST for convenience (since POST is the default from the Python client, but
        # GET is the default if using from a browser).
        cls._register_mapping(name="__steamship_dir__", verb=Verb.GET, path="/__dir__")
        cls._register_mapping(name="__steamship_dir__", verb=Verb.POST, path="/__dir__")
        end_time = time.time()
        logging.info(f"Registered package functions in {end_time - start_time} seconds.")

    def __steamship_dir__(self) -> dict:
        """Return this Invocable's PackageSpec for remote inspection -- e.g. documentation or OpenAPI generation."""
        return self._package_spec.dict(by_alias=True)

    @classmethod
    def config_cls(cls) -> Type[Config]:
        """Returns the configuration object for the Invocable.

        By default, Steamship packages and plugins will not take any configuration. Steamship packages and plugins may
        declare a configuration object which extends from Config, if needed, as follows:

        class MyPackageOrPlugin:
            class MyConfig(Config):
                ...

            @classmethod
            def config_cls(cls):
                return MyPackageOrPlugin.MyConfig
        """  # noqa: RST301
        return Config

    @classmethod
    def _register_mapping(
        cls,
        name: str,
        verb: Optional[Verb] = None,
        path: str = "",
        config: Dict[str, Union[int, float, bool, str]] = None,
    ) -> MethodSpec:
        """Registering a mapping permits the method to be invoked via HTTP."""
        method_spec = MethodSpec(cls, name, path=path, verb=verb, config=config)
        # It's important to use method_spec.path below since that's the CLEANED path.
        cls._method_mappings[verb][method_spec.path] = name
        logging.info(f"[{cls.__name__}] {verb} {path} => {name}")
        return method_spec

    def __call__(self, request: InvocableRequest, context: Any = None) -> InvocableResponse:
        """Invokes a method call if it is registered."""
        if not hasattr(self.__class__, "_method_mappings"):
            logging.error("__call__: No mappings available on invocable.")
            return InvocableResponse.error(
                code=HTTPStatus.NOT_FOUND, message="No mappings available for invocable."
            )

        if request.invocation is None:
            logging.error("__call__: No invocation on request.")
            return InvocableResponse.error(
                code=HTTPStatus.NOT_FOUND, message="No invocation was found."
            )

        verb = Verb(request.invocation.http_verb.strip().upper())
        path = request.invocation.invocation_path

        path = MethodSpec.clean_path(path)

        logging.info(f"[{verb}] {path}")

        method_mappings = self.__class__._method_mappings

        if verb not in method_mappings:
            logging.error(f"__call__: Verb '{verb}' not found in method_mappings.")
            return InvocableResponse.error(
                code=HTTPStatus.NOT_FOUND,
                message=f"No methods for verb {verb} available.",
            )

        if path not in method_mappings[verb]:
            logging.error(f"__call__: Path '{path}' not found in method_mappings[{verb}].")
            return InvocableResponse.error(
                code=HTTPStatus.NOT_FOUND,
                message=f"No handler for {verb} {path} available.",
            )

        method = method_mappings[verb][path]
        if not (hasattr(self, method) and callable(getattr(self, method))):
            logging.error(
                f"__call__: Method not found or not callable for '{path}' in method_mappings[{verb}]."
            )
            return InvocableResponse.error(
                code=HTTPStatus.INTERNAL_SERVER_ERROR,
                message=f"Handler for {verb} {path} not callable.",
            )

        arguments = request.invocation.arguments
        if arguments is None:
            return getattr(self, method)()
        else:
            return getattr(self, method)(**arguments)

    @classmethod
    def get_config_parameters(cls) -> Dict[str, ConfigParameter]:
        return cls.config_cls().get_config_parameters()



import json
from enum import Enum
from typing import Dict, List, Optional, Type, Union

from pydantic import BaseModel, StrictBool, StrictFloat, StrictInt, StrictStr

from steamship.base.error import SteamshipError


class ConfigParameterType(str, Enum):
    NUMBER = "number"
    STRING = "string"
    BOOLEAN = "boolean"

    @staticmethod
    def from_python_type(t: Type):
        if issubclass(t, str):
            return ConfigParameterType.STRING
        elif issubclass(t, bool):  # bool is a subclass of int, so must do this first!
            return ConfigParameterType.BOOLEAN
        elif issubclass(t, float) or issubclass(t, int):
            return ConfigParameterType.NUMBER
        else:
            raise SteamshipError(f"Unknown value type in Config: {t}")


class ConfigParameter(BaseModel):
    type: ConfigParameterType
    description: Optional[str] = None

    # Use strict so that Pydantic doesn't coerce values into the first one that fits
    default: Optional[Union[StrictStr, StrictBool, StrictFloat, StrictInt]] = None


class DeployableType(str, Enum):
    PLUGIN = "plugin"
    PACKAGE = "package"


class SteamshipRegistry(BaseModel):
    tagline: Optional[str]  # noqa: N815
    tagline2: Optional[str]  # noqa: N815
    usefulFor: Optional[str]  # noqa: N815
    videoUrl: Optional[str]  # noqa: N815
    githubUrl: Optional[str]  # noqa: N815
    demoUrl: Optional[str]  # noqa: N815
    blogUrl: Optional[str]  # noqa: N815
    jupyterUrl: Optional[str]  # noqa: N815
    authorGithub: Optional[str]  # noqa: N815
    authorName: Optional[str]  # noqa: N815
    authorEmail: Optional[str]  # noqa: N815
    authorTwitter: Optional[str]  # noqa: N815
    authorUrl: Optional[str]  # noqa: N815
    tags: List[str]


class PluginConfig(BaseModel):
    isTrainable: Optional[bool] = False  # noqa: N815
    transport: str = "jsonOverHttp"
    type: str  # Does not use PluginType due to circular import


class Manifest(BaseModel):
    type: DeployableType
    handle: str
    version: str
    description: Optional[str]
    author: Optional[str]
    entrypoint: str = "Unused"
    public: bool
    plugin: Optional[PluginConfig]
    build_config: Dict[str, List[str]] = {"ignore": []}
    configTemplate: Optional[Dict[str, ConfigParameter]]  # noqa: N815
    steamshipRegistry: SteamshipRegistry  # noqa: N815

    @staticmethod
    def load_manifest() -> "Manifest":
        return Manifest.parse_file("steamship.json", content_type="application/json")

    def save(self):
        with open("steamship.json", "w") as file:
            json.dump(self.dict(), file, indent="\t")

    def config_template_as_dict(self):
        result = {}
        for param, spec in self.configTemplate.items():
            result[param] = {k: v for k, v in spec.dict().items() if v is not None}

        return result



__copyright__ = "Steamship"
__license__ = "MIT"

import json
from enum import Enum

from steamship.base.model import CamelModel


class TrainingPlatform(str, Enum):
    """Note: The `str` parent object is critical for serialization!"""

    ECS = "ecs"
    LAMBDA = "lambda"


class PluginInstance(CamelModel):
    training_platform: TrainingPlatform


def test_enum_serialization_behavior():
    assert json.dumps(TrainingPlatform.ECS) == '"ecs"'

    t1 = PluginInstance.parse_obj({"training_platform": "ecs"})
    t2 = PluginInstance.parse_obj({"training_platform": "lambda"})

    assert t1.training_platform == TrainingPlatform.ECS
    assert t2.training_platform == TrainingPlatform.LAMBDA
    assert t1.training_platform == "ecs"
    assert t2.training_platform == "lambda"



import time
import webbrowser

import requests

from steamship.base.error import SteamshipError


def login(api_base: str, web_base: str) -> str:  # noqa: C901

    # create login token
    try:
        token_result = requests.post(api_base + "account/create_login_attempt")
        token_data = token_result.json().get("data")
    except Exception as e:
        raise SteamshipError("Could not create login token when attempting login.", error=e)

    if token_data is None:
        raise SteamshipError("Could not create login token when attempting login.")
    token = token_data.get("token")
    if token is None:
        raise SteamshipError("Could not create login token when attempting login.")

    # Launch login attempt in browser
    login_browser_url = (
        f"{web_base}account/client-login?attemptToken={token}&client=pycli&version=0.0.1"
    )
    try:
        opened_browser = webbrowser.open(login_browser_url)
    except Exception as e:
        raise SteamshipError("Exception attempting to launch browser for login.", error=e)

    if not opened_browser:
        raise SteamshipError(
            """Could not launch browser to log in to Steamship.

If you are in Replit:

1) Get an API key at https://steamship.com/account/api
2) Set the STEAMSHIP_API_KEY Replit Secret
3) Close and reopen this shell so that secrets refresh

If you are in a different headless environment, visit https://docs.steamship.com/configuration/authentication.html"""
        )

    # Wait on result
    total_poll_time_s = 0
    time_between_polls_s = 1
    api_key = None
    while total_poll_time_s < 300:  # Five minutes
        params = {"token": token}
        login_response = requests.post(f"{api_base}account/poll_login_attempt", json=params).json()
        if login_response.get("data", {}).get("status") == "done":
            api_key = login_response.get("data", {}).get("apiKey")
            break
        time.sleep(time_between_polls_s)
        total_poll_time_s += time_between_polls_s

    if api_key is None:
        raise SteamshipError("Could not fetch api key after login attempt in allotted time.")

    return api_key



from __future__ import annotations

from typing import Any, Type

from pydantic import BaseModel, Field

from steamship.base.client import Client
from steamship.base.model import CamelModel


class User(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    handle: str = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["user"] if "user" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def current(client: Client) -> User:
        return client.get("account/current", expect=User)



from __future__ import annotations

from typing import List, Optional

from steamship.base.request import Request
from steamship.base.response import Response
from steamship.data.block import Block


class GenerateRequest(Request):
    """This class provides the input for a request to a Generator.  There are several ways to specify the input; see below"""

    # Input Specification
    # You must select one of several ways to specify input for a generator. These are exclusive.
    # 1 - A span of Blocks on a File
    # 2 - Raw text
    # 3 - A query for Blocks
    # 4 - (coming soon) A public URL of content
    # 5 - (coming soon) Raw bytes of content

    # Must specify plugin instance to use
    plugin_instance: str = None

    # May specify blocks by their file_id. If so, may specify start and end index
    input_file_id: str = None
    input_file_start_block_index: int = None
    input_file_end_block_index: Optional[
        int
    ] = None  # EXCLUSIVE end index, like most programming languages

    # May specify raw text
    text: Optional[str] = None

    # May specify raw bytes (ex. an image, audio) [Not yet implemented]
    # bytes: Optional[bytes] = None

    # May specify a block query. This may produce input blocks from multiple files.
    block_query: Optional[str] = None

    # May specify a public URL to fetch the data from. [Not yet implemented]
    # url: Optional[str] = None

    # Desired output specification

    # Whether we want the output appended to a file
    append_output_to_file: bool = False

    # May specify a file to which to append the results.  This may be the same file as
    # the input or not.  If appendOutputToFile is true but the outputFileId is not set,
    # create a new file.
    output_file_id: Optional[str] = None

    # Arbitrary runtime options which may be passed to a generator
    options: Optional[dict]


class GenerateResponse(Response):
    blocks: List[Block]



from typing import List

import pytest
from pydantic import BaseModel
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import SteamshipError, Task, TaskState


def schedule_task(client, please_fail: bool = False, dependencies: List[Task] = None) -> Task:
    response = client.post(
        "task/alwaysfail" if please_fail else "task/noop",
        payload={},
        expect=Task[BaseModel],
        as_background_task=True,
        wait_on_tasks=dependencies,
    )
    assert response is not None
    assert response.state == TaskState.waiting
    assert response.task_id is not None
    return response


def test_task_wait_callback():
    """Test basic connection"""
    client = get_steamship_client()

    task1 = schedule_task(client)

    retries_checked = 0

    def on_task_check(retry_number, total_sec, task):
        nonlocal retries_checked
        retries_checked = retries_checked + 1
        assert retry_number
        assert retry_number > 0
        assert total_sec
        assert total_sec > 0
        assert task.task_id == task1.task_id

    task1.wait(on_each_refresh=on_task_check)
    assert retries_checked > 0
    assert task1.state == TaskState.succeeded


def test_task_dependencies_parallel_success():
    """Test basic connection"""
    client = get_steamship_client()

    task1 = schedule_task(client)
    task2 = schedule_task(client)
    task3 = schedule_task(client, dependencies=[task1, task2])

    task1.wait()
    assert task1.state == TaskState.succeeded

    task2.wait()
    assert task2.state == TaskState.succeeded

    task3.wait()
    assert task3.state == TaskState.succeeded


def test_task_dependencies_parallel_failure():
    """Test basic connection"""
    client = get_steamship_client()

    task1 = schedule_task(client)
    task2 = schedule_task(client, please_fail=True)
    task3 = schedule_task(client, dependencies=[task1, task2])

    task1.wait()
    assert task1.state == TaskState.succeeded

    with pytest.raises(SteamshipError):
        task2.wait()  # It failed!

    with pytest.raises(SteamshipError):
        task3.wait()  # It failed too -- because of the rollup!



from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import PackageInstance


def test_configurable_instance_invoke():
    greeting1 = "Hola"
    config_template = {
        "greeting": {"type": "string"},
        "snake_case_config": {"type": "string"},
        "camelCaseConfig": {"type": "string"},
        "defaultConfig": {"type": "string", "default": "defaulted"},
    }
    instance_config = {
        "greeting": greeting1,
        "snake_case_config": "hisss",
        "camelCaseConfig": "spit!",
    }
    client = get_steamship_client()
    hello_world_path = PACKAGES_PATH / "configurable_hello_world.py"

    with deploy_package(
        client,
        hello_world_path,
        version_config_template=config_template,
        instance_config=instance_config,
    ) as (package, version, instance):
        # Now let's invoke it!
        # Note: we're invoking the data at configurable_hello_world.py in the tests/assets/packages folder
        res = instance.invoke("greet")
        assert res == f"{greeting1}, Person"

        res = instance.invoke("snake")
        assert res == instance_config.get("snake_case_config")

        res = instance.invoke("camel")
        assert res == instance_config.get("camelCaseConfig")

        res = instance.invoke("defaulted")
        assert res == "defaulted"

        greeting2 = "Hallo"
        instance_config2 = {
            "greeting": greeting2,
            "snake_case_config": "hisss",
            "camelCaseConfig": "spit!",
        }
        instance2 = PackageInstance.create(
            client, package_id=package.id, package_version_id=version.id, config=instance_config2
        )

        res2 = instance2.invoke("greet")
        assert res2 == f"{greeting2}, Person"

        # Test with quick-create
        greeting3 = "Howdy"
        instance_config3 = {
            "greeting": greeting3,
            "snake_case_config": "hisss",
            "camelCaseConfig": "spit!",
        }
        instance3 = client.use(package.handle, f"{instance2.handle}-2", config=instance_config3)
        assert instance3.invoke("greet") == f"{greeting3}, Person"
        assert instance3.id != instance2.id

        # Test that quick-create reuses an instance
        instance3a = client.use(package.handle, f"{instance2.handle}-2", config=instance_config3)
        assert instance3a.id == instance3.id

        # Test instance with quick create and no handle
        greeting4 = "How Do Dee"
        instance_config4 = {
            "greeting": greeting4,
            "snake_case_config": "hisss",
            "camelCaseConfig": "spit!",
        }
        instance4 = client.use(package.handle, config=instance_config4)
        assert instance4.invoke("greet") == f"{greeting4}, Person"
        assert instance4.id != instance3.id



from enum import Enum
from typing import Optional

from steamship.invocable import InvocableResponse, PackageService, post


class DropdownValue(str, Enum):
    VALUE1 = "value1"
    VALUE2 = "value2"


class OptionalParams(PackageService):
    @post("enum_route")
    def enum_route(
        self, dropdown_value: Optional[DropdownValue] = DropdownValue.VALUE1
    ) -> InvocableResponse:
        return InvocableResponse(string=f"You picked {dropdown_value}")



from steamship.base.model import CamelModel


class Request(CamelModel):
    pass


class GetRequest(Request):
    id: str = None
    handle: str = None


class CreateRequest(Request):
    id: str = None
    handle: str = None


class UpdateRequest(Request):
    id: str = None
    handle: str = None


class IdentifierRequest(Request):
    id: str = None
    handle: str = None


class ListRequest(Request):
    pass


class DeleteRequest(Request):
    id: str



import pytest

from steamship import File, Steamship, SteamshipError
from steamship.base import TaskState


@pytest.mark.usefixtures("client")
def test_e2e_blockifier_markdown(client: Steamship):
    # TODO (enias): Stop offering default plugins hosted on the engine
    blockifier = client.use_plugin("markdown-blockifier-default")
    file = File.create(client=client, content="This is a test.")
    assert len(file.refresh().blocks) == 0
    task = file.blockify(plugin_instance=blockifier.handle)
    task.wait()

    if task.state == TaskState.failed:
        raise SteamshipError(message=task.status_message)
    assert task.state == TaskState.succeeded

    assert len(file.refresh().blocks) == 1
    file.delete()



"""CSV Blockifier - Steamship Plugin."""

import csv
import io
from typing import List, Optional, Type, Union

from pydantic import constr

from steamship.base import SteamshipError
from steamship.data.block import Block
from steamship.data.file import File
from steamship.data.tags import Tag
from steamship.invocable import Config, InvocableResponse, create_handler
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.blockifier.blockifier import Blockifier
from steamship.plugin.inputs.raw_data_plugin_input import RawDataPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput


class CsvBlockifier(Blockifier):
    """Converts CSV or TSV into Tagged Steamship Blocks.

    A CSV file is mapped onto 1 File
    Each row in the CSV file is mapped onto a block
    Each block has text which is extracted from a single column
    Each block has one or more tags extracted from the values in one ore more columns
    """

    class CsvBlockifierConfig(Config):
        text_column: str
        tag_columns: List[str]
        tag_kind: str = None
        delimiter: Optional[str] = ","
        quotechar: Optional[constr(max_length=1)] = '"'
        escapechar: Optional[constr(max_length=1)] = "\\"
        newline: Optional[str] = "\\n"
        skipinitialspace: Optional[bool] = False

        def __init__(self, **kwargs):
            if isinstance(kwargs["tag_columns"], str):
                kwargs["tag_columns"] = kwargs["tag_columns"].split(",")
            super().__init__(**kwargs)

    @classmethod
    def config_cls(cls) -> Type[CsvBlockifierConfig]:
        return cls.CsvBlockifierConfig

    def _get_text(self, row) -> str:
        text = row.get(self.config.text_column)
        if text:
            text = text.replace(self.config.newline, "\n")
        return text

    def _get_tags(self, row) -> List[str]:
        tag_values = []
        for tag_column in self.config.tag_columns:
            tag_value = row.get(tag_column)
            if tag_value:
                tag_values.append(tag_value.replace(self.config.newline, "\n"))
        return tag_values

    def run(
        self, request: PluginRequest[RawDataPluginInput]
    ) -> Union[InvocableResponse, InvocableResponse[BlockAndTagPluginOutput]]:
        if request is None or request.data is None or request.data.data is None:
            return InvocableResponse(
                error=SteamshipError(message="Missing data field on the incoming request.")
            )
        data = request.data.data  # TODO (enias): Simplify
        if isinstance(data, bytes):
            data = data.decode("utf-8")

        if not isinstance(data, str):
            return InvocableResponse(
                error=SteamshipError(message="The incoming data was not of expected String type")
            )

        reader = csv.DictReader(
            io.StringIO(data),
            delimiter=self.config.delimiter,
            quotechar=self.config.quotechar,
            escapechar=self.config.escapechar,
            skipinitialspace=self.config.skipinitialspace,
        )
        file = File(blocks=[])
        for row in reader:
            text = self._get_text(row)
            tag_values = self._get_tags(row)

            block = Block(
                text=text,
                tags=[Tag(kind=self.config.tag_kind, name=tag_value) for tag_value in tag_values],
            )
            file.blocks.append(block)

        return InvocableResponse(data=BlockAndTagPluginOutput(file=file))


handler = create_handler(CsvBlockifier)



from enum import Enum

from steamship.invocable import InvocableResponse, PackageService, longstr, post


class DropdownValue(str, Enum):
    VALUE1 = "value1"
    VALUE2 = "value2"


class FancyTypes(PackageService):
    @post("enum_route")
    def enum_route(self, dropdown_value: DropdownValue) -> InvocableResponse:
        return InvocableResponse(string=f"You picked {dropdown_value}")

    @post("long_string_route")
    def long_string_route(self, long_long_string: longstr) -> InvocableResponse:
        return InvocableResponse(string=f"You gave me a long string {long_long_string}")
